# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_value.common.tables.comparators.table_2_value_table_comparator import Table2ValueTable_Comparator
from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_rabbitmq.tools.rabbitmq import rabbitmq_client
from holado_core.common.tables.table_manager import TableManager
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.tools.tools import Tools
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
import queue
from holado_system.system.global_system import GlobalSystem
from holado_core.common.exceptions.technical_exception import TechnicalException
import pika
from holado_rabbitmq.tools.rabbitmq.rabbitmq_blocking_client import RMQBlockingClient
from holado_rabbitmq.tools.rabbitmq.rabbitmq_select_client import RMQSelectClient
from holado_python.standard_library.typing import Typing


logger = logging.getLogger(__name__)


if rabbitmq_client.RMQClient.is_available():

    def __get_session_context():
        return SessionContext.instance()
    
    def __get_scenario_context():
        return __get_session_context().get_scenario_context()
    
    def __get_text_interpreter():
        return __get_scenario_context().get_text_interpreter()
    
    def __get_variable_manager():
        return __get_scenario_context().get_variable_manager()
    
    def __get_protobuf_messages():
        return SessionContext.instance().protobuf_messages
    
    def __get_rabbitmq_manager():
        return SessionContext.instance().rabbitmq_manager
    
    
    def __convert_step_args_2_RabbitMQClient_kwargs(step_args):
        res = {}
        args = dict(step_args)
        if 'queue.name' in args:
            res['queue'] = args.pop('queue.name')
        if Tools.has_sub_kwargs(args, 'queue.'):
            res['queue_args'] = Tools.pop_sub_kwargs(args, 'queue.')
        if 'exchange.name' in args:
            res['exchange'] = args.pop('exchange.name')
        if Tools.has_sub_kwargs(args, 'exchange.'):
            res['exchange_args'] = Tools.pop_sub_kwargs(args, 'exchange.')
        if Tools.has_sub_kwargs(args, 'bind.'):
            res['bind_args'] = Tools.pop_sub_kwargs(args, 'bind.')
        res.update(args)
        return res
    
    
    @Given(r"(?P<var_name>{Variable}) = new RabbitMQ client")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = RMQBlockingClient()
        __get_variable_manager().register_variable(var_name, obj)
    
    @Given(r"(?P<var_name>{Variable}) = new RabbitMQ client with rapid close")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = RMQBlockingClient()
        obj.rapid_close = True
        __get_variable_manager().register_variable(var_name, obj)
    
    @Given(r"(?P<var_name>{Variable}) = new asynchronous RabbitMQ client")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj = RMQSelectClient()
        __get_variable_manager().register_variable(var_name, obj)
    
    @When(r"connect \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Create a new connection to RabbitMQ Client. Connection parameters are defined in a table.
        The connection is a select connection for asynchronous client else a blocking connection.
        Credentials:
            To define connection credentials, define username and password values
        """

        if hasattr(context, "table"):
            table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
            connection_parameters_kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
        else:
            connection_parameters_kwargs = {}
        
        client = StepTools.evaluate_variable_value(var_client)
        client.connect(**connection_parameters_kwargs)
    
    @When(r"connect with a blocking connection \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Create a new blocking connection to RabbitMQ Client. Connection parameters are defined in a table.
        Credentials:
            To define connection credentials, define username and password values
        """
        if hasattr(context, "table"):
            table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
            connection_parameters_kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
        else:
            connection_parameters_kwargs = {}
        
        client = StepTools.evaluate_variable_value(var_client)
        if not isinstance(client, RMQBlockingClient):
            raise FunctionalException(f"A blocking connection is only possible with a blocking client")
        client.connect(**connection_parameters_kwargs)
    
    @When(r"connect with a select connection \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Create a new select connection to RabbitMQ Client. 
        Connection parameters are defined in a table.
        A parameter 'wait_connection' can be set to wait connection before step end (default).
        Credentials:
            To define connection credentials, define username and password values
        """

        if hasattr(context, "table"):
            table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
            connection_parameters_kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
        else:
            connection_parameters_kwargs = {}
        
        client = StepTools.evaluate_variable_value(var_client)
        if not isinstance(client, RMQBlockingClient):
            raise FunctionalException(f"A select connection is only possible with a select client")
        client.connect(**connection_parameters_kwargs)
    
    @Step(r"flush \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Flush connection of RabbitMQ Client
        """
        client = StepTools.evaluate_variable_value(var_client)
        client.flush()
    
    @Step(r"close \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Close connection to RabbitMQ Client
        """
        client = StepTools.evaluate_variable_value(var_client)
        client.close()
    
    @Given(r"(?P<var_name>{Variable}) = new publisher on queue (?P<queue>{Str}) \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, queue, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        queue = StepTools.evaluate_scenario_parameter(queue)
        
        execute_steps("""
            Given {} = new publisher (RMQ client: {})
                | Name         | Value |
                | 'queue.name' | '{}'  |
            """.format(var_name, var_client, queue) )
    
    @Given(r"(?P<var_name>{Variable}) = new publisher \(RMQ client: (?P<var_client>{Variable})(?: ; nb of runners: (?P<nb_runners>{Int}))?\)")
    def step_impl(context, var_name, var_client, nb_runners):
        var_name = StepTools.evaluate_variable_name(var_name)
        nb_runners = StepTools.evaluate_scenario_parameter(nb_runners)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        
        args = ValueTableConverter.convert_name_value_table_2_dict(table)
        kwargs = __convert_step_args_2_RabbitMQClient_kwargs(args)
        if nb_runners is not None and 'nb_runners' not in kwargs:
            kwargs['nb_runners'] = nb_runners
        
        client = StepTools.evaluate_variable_value(var_client)
        pub = client.new_publisher(**kwargs)
        
        __get_variable_manager().register_variable(var_name, pub)
    
    @Given(r"(?P<var_name>{Variable}) = new buffer consumer on queue (?P<queue>{Str}) \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, queue, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        queue = StepTools.evaluate_scenario_parameter(queue)
        
        execute_steps("""
            Given {} = new buffer consumer (RMQ client: {})
                | Name         | Value |
                | 'queue.name' | '{}'  |
            """.format(var_name, var_client, queue) )
    
    @Given(r"(?P<var_name>{Variable}) = new buffer consumer \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        args = ValueTableConverter.convert_name_value_table_2_dict(table)
        kwargs = __convert_step_args_2_RabbitMQClient_kwargs(args)
        
        client = StepTools.evaluate_variable_value(var_client)
        con = client.new_buffer_consumer(**kwargs)
        
        __get_variable_manager().register_variable(var_name, con)
    
    # @Given(r"start consuming in a thread \(RMQ client: (?P<var_client>{Variable})\)")
    # def step_impl(context, var_client):
    #     # execute_steps(u"""
    #     #     Given __THREAD_ID_CONSUMING__@ = start consuming in a thread (RMQ client: {var_client})
    #     #     Given wait until client is consuming (RMQ client: {var_client}) (accepted time: 20 s)
    #     #     Given for thread __THREAD_ID_CONSUMING__@, call steps for interrupt
    #     #         \"""
    #     #         When stop consuming (RMQ client: {var_client})
    #     #         \"""
    #     #     """.format(var_client=var_client))
    #
    #     # The wait of 0.01 seconds is done to avoid the case where two execute_steps are done in the same time in two different threads
    #     execute_steps(u"""
    #         Given __THREAD_ID_CONSUMING__@ = start consuming in a thread (RMQ client: {var_client})
    #         Given wait 0.01 seconds
    #         Given wait until client is consuming (RMQ client: {var_client}) (accepted time: 5 s)
    #         """.format(var_client=var_client))
    #
    # @Given(r"(?P<var_name>{Variable}) = start consuming in a thread \(RMQ client: (?P<var_client>{Variable})\)")
    # def step_impl(context, var_name, var_client):
    #     var_name = StepTools.evaluate_variable_name(var_name)
    #     var_client = StepTools.evaluate_variable_name(var_client)
    #
    #     # Start consuming in a small delay
    #     execute_steps(u"""
    #         Given {var_name} = call steps in a thread
    #             \"""
    #             When start consuming (RMQ client: {var_client})
    #             \"""
    #         """.format(var_name=var_name, var_client=var_client))
    #
    #     # Register the thread allowed to start consuming 
    #     thread_name = __get_variable_manager().get_variable_value(var_name)
    #     thread = SessionContext.instance().threads_manager.get_thread(thread_name)
    #     thread_id = MultitaskManager.get_thread_id(thread=thread)
    #
    #     client = StepTools.evaluate_variable_value(var_client)
    #     client.consuming_thread_id = thread_id
    
    @Step(r"start consuming \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        client = StepTools.evaluate_variable_value(var_client)
        client.start_consuming()
    
    @Then(r"client is consuming \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        client = StepTools.evaluate_variable_value(var_client)
        if not client.is_consuming:
            raise FunctionalException(f"RMQ client is not consuming")
        
    @Step(r"stop consuming \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        client = StepTools.evaluate_variable_value(var_client)
        client.stop_consuming()
    
    @Step(r"start data events processing \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Usually this step is not needed, as data events processing is automatically started at first publish.
        """
        client = StepTools.evaluate_variable_value(var_client)
        client.start_process_data_events()
    
    @Step(r"stop data events processing \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_client):
        """
        Usually this step is not needed, as data events processing is automatically stopped when closing client.
        """
        client = StepTools.evaluate_variable_value(var_client)
        client.stop_process_data_events()
    
    @When(r"publish (?P<msg>{Str}) \(RMQ publisher: (?P<var_pub>{Variable})\)")
    def step_impl(context, msg, var_pub):
        msg = StepTools.evaluate_scenario_parameter(msg)
        pub = StepTools.evaluate_variable_value(var_pub)
        pub.publish(msg)
    
    @When(r"publish all of (?P<iterable_str>{Any}) \(RMQ publisher: (?P<var_pub>{Variable})\)")
    def step_impl(context, iterable_str, var_pub):
        iterable = StepTools.evaluate_scenario_parameter(iterable_str)
        pub = StepTools.evaluate_variable_value(var_pub)
        pub_name = StepTools.evaluate_variable_name(var_pub)
        
        counter = 0
        
        for msg in iterable:
            counter += 1
            
            try:
                pub.publish(msg)
            except pika.exceptions.ChannelWrongStateError as exc:
                raise TechnicalException(f"[Publisher {pub_name}] Stop publishing after error on message {counter}: [{Typing.get_object_class_fullname(exc)}] {str(exc)}") from exc
            except Exception as exc:
                raise TechnicalException(f"[Publisher {pub_name}] Failed to publish message {counter}: [{Typing.get_object_class_fullname(exc)}] {str(exc)}") from exc
                
            # Yield processor in case of multithreading
            if counter % 10 == 0:
                GlobalSystem.yield_processor()
    
    @When(r"publish all of queue (?P<var_queue>{Variable}) \(RMQ publisher: (?P<var_pub>{Variable})(?: ; rate log period: (?P<rate_period_s>{Int}) s)?\)")
    def step_impl(context, var_queue, var_pub, rate_period_s):
        q = StepTools.evaluate_variable_value(var_queue)
        q_name = StepTools.evaluate_variable_name(var_queue)
        pub = StepTools.evaluate_variable_value(var_pub)
        pub_name = StepTools.evaluate_variable_name(var_pub)
        rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
        
        counter = 0
        beg = Tools.timer_s()
        c_beg, t_beg = counter, beg
        log_counter = 0
        
        try:
            while True:
                # Get next message and publish
                msg = q.get()
                if q.is_sentinel(msg):
                    break
                counter += 1
                
                try:
                    pub.publish(msg)
                except pika.exceptions.ChannelWrongStateError as exc:
                    logger.error(f"[Queue {q_name} -> Publisher {pub_name}] Stop publishing after error on message {counter}: [{Typing.get_object_class_fullname(exc)}] {str(exc)}")
                    break
                except Exception as exc:
                    logger.warning(f"[Queue {q_name} -> Publisher {pub_name}] Failed to publish message {counter}: [{Typing.get_object_class_fullname(exc)}] {str(exc)}")
                    # logger.warning(f"[Queue {q_name} -> Publisher {pub_name}] Failed to publish message {counter}: [{Typing.get_object_class_fullname(exc)}] {Tools.represent_exception(exc)}")
                finally:
                    q.task_done()
                
                # Log rate if needed
                if rate_period_s and counter % 10 == 0:
                    t_end = Tools.timer_s()
                    if t_end > t_beg + rate_period_s:
                        c_end, s_end = counter, q.qsize()
                        rate = (c_end - c_beg) / (t_end - t_beg)
                        log_counter += 1
                        logger.print(f"[Queue {q_name} -> Publisher {pub_name}] Rate: {int(rate)} msg/s ; Nb messages: {counter} ; Queue size: {s_end}")
                        c_beg, t_beg = c_end, t_end
                        
                # Yield processor in case of multithreading
                if counter % 10 == 0:
                    GlobalSystem.yield_processor()
        except queue.Empty:
            # Without block, or with timeout, this exception occurs when queue is empty
            pass
        
        if rate_period_s:
            end = Tools.timer_s()
            rate = counter / (end - beg)
            logger.print(f"[Queue {q_name} -> Publisher {pub_name}] Mean rate: {int(rate)} msg/s ; Nb messages: {counter}")
    
    @When(r"await consumer (?P<var_con>{Variable}) receives a message(?: \(timeout: (?P<timeout_sec>{Float}) s(?: ; polling: (?P<polling_sec>{Float}) s)?\))?")
    def step_impl(context, var_con, timeout_sec, polling_sec):
        con = StepTools.evaluate_variable_value(var_con)
        timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
        polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
        
        __get_rabbitmq_manager().await_message_reception([con], timeout_seconds=timeout_sec, polling_seconds=polling_sec)
    
    @When(r"await consumers (?P<list_consumers>{List}) receive a message(?: \(timeout: (?P<timeout_sec>{Float}) s(?: ; polling: (?P<polling_sec>{Float}) s)?\))?")
    def step_impl(context, list_consumers, timeout_sec, polling_sec):
        consumers = StepTools.evaluate_list_scenario_parameter(list_consumers, "list_consumers")
        timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
        polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
        
        __get_rabbitmq_manager().await_message_reception(consumers, timeout_seconds=timeout_sec, polling_seconds=polling_sec)
    
    @When(r"await consumer (?P<var_con>{Variable}) receives (?:(?P<no_str>no) )?messages(?: \((?:first timeout: (?P<first_timeout_sec>{Float}) s)?(?: ; )?(?:window: (?P<window_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
    def step_impl(context, var_con, no_str, first_timeout_sec, window_sec, polling_sec):
        con = StepTools.evaluate_variable_value(var_con)
        first_timeout_sec = StepTools.evaluate_scenario_parameter(first_timeout_sec)
        window_sec = StepTools.evaluate_scenario_parameter(window_sec)
        polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
        do_raise = no_str is None
        
        __get_rabbitmq_manager().await_messages_reception([con], first_timeout_seconds=first_timeout_sec, window_seconds=window_sec, polling_seconds=polling_sec, raise_exception=do_raise)
    
    @When(r"await consumers (?P<list_consumers>{List}) receive (?:(?P<no_str>no) )?messages(?: \((?:first timeout: (?P<first_timeout_sec>{Float}) s)?(?: ; )?(?:window: (?P<window_sec>{Float}) s)?(?: ; )?(?:polling: (?P<polling_sec>{Float}) s)?\))?")
    def step_impl(context, list_consumers, no_str, first_timeout_sec, window_sec, polling_sec):
        consumers = StepTools.evaluate_list_scenario_parameter(list_consumers, "list_consumers")
        first_timeout_sec = StepTools.evaluate_scenario_parameter(first_timeout_sec)
        window_sec = StepTools.evaluate_scenario_parameter(window_sec)
        polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
        do_raise = no_str is None
        
        __get_rabbitmq_manager().await_messages_reception(consumers, first_timeout_seconds=first_timeout_sec, window_seconds=window_sec, polling_seconds=polling_sec, raise_exception=do_raise)
    
    @Then(r"consumer (?P<var_con>{Variable}) doesn't receive any message \(timeout: (?P<timeout_sec>{Float}) s ; polling: (?P<polling_sec>{Float}) s\)")
    def step_impl(context, var_con, timeout_sec, polling_sec):
        con = StepTools.evaluate_variable_value(var_con)
        timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
        polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
        
        nb = __get_rabbitmq_manager().await_message_reception([con], timeout_seconds=timeout_sec, polling_seconds=polling_sec, raise_exception=False)
        if nb > 0:
            raise FunctionalException(f"Consumer has received {nb} messages")
    
    @Then(r"consumers (?P<list_consumers>{List}) don't receive any message \(timeout: (?P<timeout_sec>{Float}) s ; polling: (?P<polling_sec>{Float}) s\)")
    def step_impl(context, list_consumers, timeout_sec, polling_sec):
        consumers = StepTools.evaluate_list_scenario_parameter(list_consumers, "list_consumers")
        timeout_sec = StepTools.evaluate_scenario_parameter(timeout_sec)
        polling_sec = StepTools.evaluate_scenario_parameter(polling_sec)
        
        nb = __get_rabbitmq_manager().await_message_reception(consumers, timeout_seconds=timeout_sec, polling_seconds=polling_sec, raise_exception=False)
        if nb > 0:
            raise FunctionalException(f"Consumer has received {nb} messages")
    
    @Then(r"consumer (?P<var_con>{Variable}) received (?P<nb_msg>{Int}) messages")
    def step_impl(context, var_con, nb_msg):
        con = StepTools.evaluate_variable_value(var_con)
        nb_msg = StepTools.evaluate_scenario_parameter(nb_msg)
    
        if len(con.messages) != nb_msg:
            raise FunctionalException(f"The number of received messages is {len(con.messages)} (expected: {nb_msg}). Received messages:\n    {(chr(10)+'    ').join((str(m[3]) for m in con.messages))}")
    
    @Then(r"consumer (?P<var_con>{Variable}) received(?: (?P<nb_msg>{Int}))? string messages:")
    def step_impl(context, var_con, nb_msg):
        con = StepTools.evaluate_variable_value(var_con)
        nb_msg = StepTools.evaluate_scenario_parameter(nb_msg)
        table = BehaveStepTools.convert_step_table_2_value_table(context.table)
        
        # Verify number of messages
        if nb_msg:
            if len(con.messages) != nb_msg:
                raise FunctionalException(f"The number of received messages is {len(con.messages)} (expected: {nb_msg}). Received messages:\n    {(chr(10)+'    ').join((str(m[3]) for m in con.messages))}")
        
        # Verify messages
        messages = [m[3].decode('utf-8') for m in con.messages]
        obtained = TableManager.convert_list_2_column_table(messages)
        comparator = Table2ValueTable_Comparator()
        comparator.equals(obtained, table)
    
    @Step(r"(?P<var_name>{Variable}) = number of received messages \(RMQ consumer: (?P<var_con>{Variable})\)")
    def step_impl(context, var_name, var_con):
        var_name = StepTools.evaluate_variable_name(var_name)
        con = StepTools.evaluate_variable_value(var_con)
        
        res = len(con.messages)

        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = received messages \(RMQ consumer: (?P<var_con>{Variable})\)")
    def step_impl(context, var_name, var_con):
        var_name = StepTools.evaluate_variable_name(var_name)
        con = StepTools.evaluate_variable_value(var_con)
        
        res = [m[3] for m in con.messages]
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Received {len(res)} messages")

        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = received messages as (?P<proto_type_str>{Str}) Protobuf objects \(RMQ consumer: (?P<var_con>{Variable})\)")
    def step_impl(context, var_name, proto_type_str, var_con):
        var_name = StepTools.evaluate_variable_name(var_name)
        proto_type_str = StepTools.evaluate_scenario_parameter(proto_type_str)
        con = StepTools.evaluate_variable_value(var_con)
        
        res = __get_rabbitmq_manager().received_messages_as_protobuf_objects(con, proto_type_str)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Step(r"(?P<var_name>{Variable}) = pop first received message \(RMQ consumer: (?P<var_con>{Variable})\)")
    def step_impl(context, var_name, var_con):
        var_name = StepTools.evaluate_variable_name(var_name)
        con = StepTools.evaluate_variable_value(var_con)
        
        message = con.pop_first_message()
        res = message[3]
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Step(r"(?P<var_name>{Variable}) = pop first received message as (?P<proto_type_str>{Str}) Protobuf object \(RMQ consumer: (?P<var_con>{Variable})\)")
    def step_impl(context, var_name, proto_type_str, var_con):
        var_name = StepTools.evaluate_variable_name(var_name)
        con = StepTools.evaluate_variable_value(var_con)
        
        proto_type_str = StepTools.evaluate_scenario_parameter(proto_type_str)
        
        message = con.pop_first_message()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Pop first message: {message[3]}")
        res = __get_session_context().protobuf_messages.new_message(proto_type_str, serialized_string=message[3])
        # logger.debug(f"+++++ New message: type={Typing.get_object_class_fullname(msg)} ; dir={dir(msg)}")
        
        __get_variable_manager().register_variable(var_name, res)
        
    @When(r"(?P<var_name>{Variable}) = pop and save received messages \(RMQ consumer: (?P<var_con>{Variable}) ; file: (?P<var_file>{Variable})(?: ; max messages: (?P<max_messages>{Int}))?\)")
    def step_impl(context, var_name, var_con, var_file, max_messages):
        var_name = StepTools.evaluate_variable_name(var_name)
        con = StepTools.evaluate_variable_value(var_con)
        file = StepTools.evaluate_variable_value(var_file)
        max_messages = int(max_messages) if max_messages else None
        
        res = con.pop_and_save_messages(file, max_messages)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Given(r"reset stored messages in consumer (?P<var_con>{Variable})")
    def step_impl(context, var_con):
        con = StepTools.evaluate_variable_value(var_con)
        con.reset_messages()
        
    @Given(r"reset stored messages in consumers (?P<list_consumers>{List})")
    def step_impl(context, list_consumers):
        consumers = StepTools.evaluate_list_scenario_parameter(list_consumers, "list_consumers")
        for con in consumers:
            con.reset_messages()
        
    @When(r"purge queue (?P<queue>{Str}) \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, queue, var_client):
        queue = StepTools.evaluate_scenario_parameter(queue)
        client = StepTools.evaluate_variable_value(var_client)
        client.purge_queue(queue)
        
    @Step(r"delete and recreate stream queue (?P<queue>{Str}) \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, queue, var_client):
        queue_name = StepTools.evaluate_scenario_parameter(queue)
        client = StepTools.evaluate_variable_value(var_client)
        client.delete_queue(queue_name)
        client._queue_declare(
            channel=client.connection.channel(),
            queue=queue_name,
            queue_args={
                'durable': True,
                "x-queue-type": "stream",
                "x-stream-max-segment-size-bytes": 100000000
            })
        
    @Step(r"bind queue (?P<queue>{Str}) to exchange (?P<exchange>{Str})(?: with routine key (?P<routine_key>{Str}))? \(RMQ client: (?P<var_client>{Variable})\)")
    def step_impl(context, queue, exchange, routine_key, var_client):
        queue_name = StepTools.evaluate_scenario_parameter(queue)
        exchange = StepTools.evaluate_scenario_parameter(exchange)
        routine_key = StepTools.evaluate_scenario_parameter(routine_key)
        client = StepTools.evaluate_variable_value(var_client)
        bind_args={
            'routing_key': routine_key
            }
        client._queue_bind(
            channel=client.connection.channel(),
            queue=queue_name,
            exchange=exchange,
            bind_args=bind_args
        )
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from holado_core.common.tables.converters.table_converter import TableConverter
from holado_ui_selenium.ui.gui.selenium.handlers.selenium_holder import SeleniumHolder
from holado_ui_selenium.ui.gui.selenium.inspectors.selenium_inspector import SeleniumInspector
import logging
from holado_ui_selenium.ui.gui.selenium.tables.selenium_table_with_header import SeleniumTableWithHeader
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_ui_selenium.ui.gui.selenium.tables.selenium_table import SeleniumTable
from holado_ui_selenium.ui.gui.selenium.tables.selenium_table_row import SeleniumTableRow
from holado_ui_selenium.ui.gui.selenium.tables.selenium_table_cell import SeleniumTableCell
from holado_core.common.tables.enums import MergeTypes
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class SeleniumTableConverter(TableConverter):
    
    
    @classmethod
    def convert_element_to_selenium_table_with_header(cls, inspector:SeleniumInspector, table_element:SeleniumHolder):
        """
        Convert a SeleniumHolder to a table with header.
        Note: invisible cells will also be added to table
        @param inspector HTML inspector
        @param table_element SeleniumHolder corresponding to the table to convert
        @return Element as a Table
        """
        find_parameters = table_element.find_info.find_parameters.with_visibility(None)   # Visibility is None to extract also hidden rows and cells
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("Converting SeleniumHolder to SeleniumTableWithHeader...")

        res = SeleniumTableWithHeader(inspector.driver, table_element.element)
        res.parent = inspector
        
        # Get header and rows
        finder = inspector.get_finder_table_row()
        rows = finder.find_all_in(table_element, find_parameters) 
        
        # Set header
        if len(rows) > 0:
            row = rows[0]
            row.description = "header"
            
            finder = inspector.get_finder_table_cell()
            
            row_cells = finder.find_all_in(row, find_parameters)
            header = cls.convert_element_to_selenium_table_row(inspector, row, row_cells)
            res.header = header
        else:
            raise TechnicalException("Unable to find table header")
        
        # Set body
        cls.__fill_body(inspector, res, rows, 1, table_element)

        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converted SeleniumHolder to SeleniumTableWithHeader of size (nb rows:{res.nb_rows} ; nb cols:{res.nb_columns})")
        
        return res
    
    @classmethod
    def convert_element_to_selenium_table(cls, inspector:SeleniumInspector, table_element:SeleniumHolder):
        """
        Convert a SeleniumHolder to a table of SeleniumHolder.
        Note: invisible rows and cells will also be added to table
        @param inspector Inspector instance
        @param table_element Table web element
        @return Table of web elements
        """
        res = SeleniumTable(inspector.driver, table_element)
        res.parent = inspector

        sub_find_parameters = table_element.find_info.find_parameters.with_visibility(None)   # Visibility is None to extract also hidden rows
        
        # Get rows
        finder = inspector.get_finder_table_row()
        rows = finder.find_all_in(table_element, sub_find_parameters) 

        # Fill body
        cls.__fill_body(inspector, res, rows, 0, table_element)

        return res
    

    @classmethod
    def __fill_body(cls, inspector, table, rows, start_index, table_element):
        # Init
        previous_row = None
        top_merge_counters = []
        finder = inspector.get_finder_table_cell()

        # Iterate through rows and add cells to matrix
        for row_element in rows[start_index:]:
#            if (inspector.internal_api.is_displayed(row_element)):
            row_cells = finder.find_all_in(row_element, row_element.find_info.find_parameters)

            row = cls.convert_element_to_selenium_table_row(inspector, row_element, row_cells, previous_row, top_merge_counters)

            table.add_row(row)
            previous_row = row
    
    @classmethod
    def convert_element_to_selenium_table_row(cls, inspector:SeleniumInspector, row_element:SeleniumHolder, row_elements):
        """
        @param inspector Inspector instance
        @param row_element Row web element
        @param row_elements List of web elements
        @return Web element table row
        """
        res = SeleniumTableRow(inspector.driver, row_element)
        ind_cell = 0
        ind_cell_in_row = 0
        colspan_counter = 0
        
        while ind_cell_in_row < len(row_elements) or colspan_counter > 0:
            if colspan_counter > 0:
                # Add new cell merged with left cell
                new_cell = SeleniumTableCell(inspector.driver, MergeTypes.LeftMerged, res[ind_cell - 1])
                # Update counter
                colspan_counter -= 1
            else:
                cell = row_elements[ind_cell_in_row]
                
                # Manage merge cells
                cell_element = cell.element
                tag = cell_element.tag_name
                if tag.lower() == "td":
                    colspan = cell_element.get_attribute("colspan")
                    if colspan is not None:
                        span_int = Converter.to_integer(colspan)
                        if span_int > 1:
                            colspan_counter = span_int - 1
                
                new_cell = SeleniumTableCell(inspector.driver, cell)
                ind_cell_in_row += 1
            
            # Add new cell
            res.add(new_cell)
            ind_cell += 1
        
        return res
    
    @classmethod
    def convert_element_to_selenium_table_row(cls, inspector:SeleniumInspector, row_element:SeleniumHolder, row_elements, previous_row:SeleniumTableRow, top_merge_counters):
        """
        @param inspector Inspector instance
        @param row_element Row web element
        @param row_elements List of web elements
        @param previous_row Previous row converted
        @param top_merge_counters Counters of number time each row cell must be merged with top cell (In/Out)
        @return Web element table row
        """
        res = SeleniumTableRow(inspector.driver, row_element)
        colspan_counter = 0
        stop = False
        
        # Add cells according top_merge_counters
        ind_cell_in_row = 0
        ind_cell = 0
        while not stop:
            # Manage merge from previous rows
            if ind_cell >= len(top_merge_counters):
                top_merge_counters.extend([0] * (ind_cell - len(top_merge_counters) + 1))
            cell_counter = top_merge_counters[ind_cell]
            
            if cell_counter > 0:
                # Add new cell merged with top cell
                new_cell = SeleniumTableCell(inspector.driver, MergeTypes.TopMerged, previous_row[ind_cell])
                res.add(new_cell)
                # Update counter for next row
                cell_counter -= 1
                top_merge_counters.insert(ind_cell, cell_counter)
                ind_cell += 1
            elif colspan_counter > 0:
                # Add new cell merged with left cell
                new_cell = SeleniumTableCell(inspector.driver, MergeTypes.LeftMerged, res[ind_cell - 1])
                res.add(new_cell)
                # Update counter
                colspan_counter -= 1
                ind_cell += 1
            elif ind_cell_in_row < len(row_elements):
                cell = row_elements[ind_cell_in_row]
                cell_element = cell.element

                # Manage merge cells
                tag = cell_element.tag_name
                if (tag.lower() == "td"):
                    rowspan = cell_element.get_attribute("rowspan")
                    colspan = cell_element.get_attribute("colspan")
                    if rowspan is not None:
                        span_int = Converter.to_integer(rowspan)
                        if span_int > 1:
                            top_merge_counters.insert(ind_cell, span_int - 1)
                    
                    if colspan is not None:
                        span_int = Converter.to_integer(colspan)
                        if span_int > 1:
                            colspan_counter = span_int - 1
                
                # Add new cell
                res.add(SeleniumTableCell(inspector.driver, cell))
                ind_cell += 1
                ind_cell_in_row  += 1
            else:
                stop = True
        
        return res
    
    # @classmethod
    # def convert_selenium_table_to_value_table(cls, SeleniumTableWithHeader table, SeleniumInspector seleniumInspector, TableConvertParameters parameters):
    #     """
    #     Convert SeleniumHolder Tables to Scenario Tables
    #     @param table table to convert
    #     @param seleniumInspector Inspector instance
    #     @param parameters Convert parameters
    #     @return Scenario table with header
    #     """
    #     InfoTableWithHeader res = InfoTableWithHeader(None)
    #     InfoTableRow resRow
    #     SeleniumTableRow row
    #
    #     TableConvertContext context = TableConvertContext.getDefault()
    #             .withColumnIndexesByName(table.columnIndexesByHeaderStringContent())
    #
    #     # Set header
    #     TableConvertParameters headerParameters = (parameters is None ? None : parameters.withoutPatternConvertByColumn())
    #     StringTableRow header = TableConverter.convertTableRowToStringTableRow(table.getHeader(), headerParameters)
    #     header.updateAfterCreate()
    #     res.header = header
    #
    #     # Set rows
    #     for (int i = 0  i < table.nbRows()  i += 1):
    #         row = table.getRow(i)
    #         resRow = convertSeleniumTableRowToScenarioTableRow(row, seleniumInspector, context, parameters)
    #         res.addRow(resRow)
    #         resRow.updateAfterCreate(res, i)
    #
    #
    #     return res
    #
    #
    #
    # @classmethod
    # def convertSeleniumTableRowToScenarioTableRow(cls, SeleniumTableRow row, SeleniumInspector seleniumInspector, TableConvertContext context, TableConvertParameters parameters):
    #     """
    #     Convert a table row to Scenario row
    #     @param row Row to convert
    #     @param seleniumInspector Inspector instance
    #     @param context Convert context
    #     @param parameters Convert parameters
    #     @return Scenario row
    #     """
    #     InfoTableRow res = InfoTableRow(None)
    #     SeleniumHolder obtainedElement 
    #
    #     for (SeleniumTableCell cell : row):
    #         # Manage merge type
    #         if (cell.getMergeType() is not None):
    #             res.add(InfoTableCell(cell.getMergeType(), None))
    #             continue
    #
    #
    #         # Get cell content
    #         obtainedElement = (SeleniumHolder) cell.getContent()
    #         if (obtainedElement is None)
    #             raise FunctionalException("Element content is not defined.")
    #
    #         # Else string content
    #         res.add(InfoTableCell(cell.getStringContent()))
    #
    #
    #     return res



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.finders.element_finder import ElementFinder
from holado_core.common.finders.tools.find_context import ContainerFindContext
from holado_core.common.finders.tools.find_parameters import FindParameters
from holado_core.common.finders.tools.enums import FindType

logger = logging.getLogger(__name__)


class TreeFinder(ElementFinder):
    """ Generic finder in a tree of elements.
    """
    
    def __init__(self, description=None):
        super().__init__(description)
    
    def find_in_children(self, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Element found in children.
        """
        return self.find(FindType.InChildren, container, candidates, find_context, find_parameters)
    
    def find_all_in_children(self, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        return self.find_all(FindType.InChildren, container, candidates, find_context, find_parameters)
    
    def find_in_parents(self, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Element found in parents.
        """
        return self.find(FindType.InParents, container, candidates, find_context, find_parameters)
    
    def find_all_in_parents(self, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        return self.find_all(FindType.InParents, container, candidates, find_context, find_parameters)
    
    def _find_all_container(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        if find_context.find_type == FindType.InChildren:
            return self._find_all_in_children(find_context, find_parameters)
        elif find_context.find_type == FindType.InParents:
            return self._find_all_in_parents(find_context, find_parameters)
        else:
            return super()._find_all_container(find_context, find_parameters)
    
    # def _find_in_children(self, find_context:ContainerFindContext, find_parameters:FindParameters):
    #     """
    #     Find element in children of given container. 
    #     @param find_context Find context
    #     @param find_parameters Find parameters
    #     @return Element found.
    #     """
    #     if find_context.find_type == FindType.InChildren:
    #         # Unexpected case, self method should be overridden
    #         raise NotImplementedError(f"[{self.finder_description}] {self}")
    #     else:
    #         return self.find(find_context.with_find_type(FindType.InChildren), find_parameters)
    
    
    def _find_all_in_children(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        """
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found in container children.
        """
        res = []
        candidates = self._find_children(find_context, find_parameters)
        for cand in candidates:
            if self.is_valid_output(cand, find_context, find_parameters):
                res.append(cand)
        return res
    
    def _find_children(self, find_context:ContainerFindContext, find_parameters:FindParameters, container=None):
        return self.get_finder_children(find_context, find_parameters).find_all_in(container=container, find_context=find_context, find_parameters=find_parameters)
    
    def _get_finder_children(self, find_context, find_parameters):
        return self.inspector.get_finder_children(find_context, find_parameters)
    
    
    # def _find_in_parents(self, find_context:ContainerFindContext, find_parameters:FindParameters):
    #     """
    #     Find element in parents of given container. 
    #     @param find_context Find context
    #     @param find_parameters Find parameters
    #     @return Element found.
    #     """
    #     if find_context.find_type == FindType.InParents:
    #         # Unexpected case, self method should be overridden
    #         raise NotImplementedError(f"[{self.finder_description}] {self}")
    #     else:
    #         return self.find(find_context.with_find_type(FindType.InParents), find_parameters)
    
    def _find_all_in_parents(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        """
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found in container parents.
        """
        res = []
        cand = self._find_parent(find_context, find_parameters)
        if self.is_valid_output(cand, find_context, find_parameters):
            res.append(cand)
        return res
    
    def _find_parent(self, find_context:ContainerFindContext, find_parameters:FindParameters, container=None):
        return self.get_finder_parent(find_context, find_parameters).find(container=container, find_context=find_context, find_parameters=find_parameters)
    
    def _get_finder_parent(self, find_context, find_parameters):
        return self.inspector.get_finder_parent(find_context, find_parameters)
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.exceptions.technical_exception import TechnicalException
from typing import NamedTuple, Iterable
from holado_core.common.tables.table_row import TableRow
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_binary.ipc.binary import Binary
from holado_python.standard_library.typing import Typing
from holado_core.common.tools.string_tools import StrTools

logger = logging.getLogger(__name__)


class BitSeries():
    """
    Define a bit-series as a list of bit sections. 
    A bit section is a tuple (name, bit length, type, value).
    
    It can be used in 'declaration' mode, ie when bit sections are added without "value" and values are filled after declaration with method "from_hex".
    
    BitSeries can be compared. If a bit section mustn't be compared, simply set its value to None in one BitSeries.
    """
    def __init__(self, bit_sections_list = None):
        self.__series = []
        self.__index_by_name = {}
        self.__len = None
        
        self.add_bit_section(bit_sections_list = bit_sections_list)
        
    def __iter__(self):
        return self.__series.__iter__()
    
    def __next__(self):
        return self.__series.__next__()
            
    def __len__(self):
        if self.__len is None:
            self.__len = sum(iter(bs.length for bs in self.__series))
        return self.__len
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.get_bit_section(index=key).value
        elif isinstance(key, str):
            return self.get_bit_section(name=key).value
        else:
            raise TechnicalException(f"Unmanaged key: {key} (type: {Typing.get_object_class_fullname(key)})")
        
    def __setitem__(self, key, value):
        if isinstance(key, int):
            return self.set_bit_section_value(index=key, value=value)
        elif isinstance(key, str):
            return self.set_bit_section_value(name=key, value=value)
        else:
            raise TechnicalException(f"Unmanaged key: {key}")
        
    @property
    def nb_bits(self):
        return len(self)
    
    @property
    def nb_bit_sections(self):
        return len(self.__series)
    
    def add_bit_section(self, bit_section=None, bit_sections_list=None):
        """
        Parameter "bit_section" is an iterable of 3 elements (name, bit length, type) or 4 elements (name, bit length, type, value).
            The "value" is usually omitted when in declaration mode, ie when .
        Parameter "bit_sections_list" is an iterable of "bit_section".
        """
        if bit_sections_list is not None:
            if not isinstance(bit_sections_list, Iterable):
                raise TechnicalException(f"Parameter 'bit_sections_list' has to be an iterable (obtained type: {Typing.get_object_class_fullname(bit_sections_list)})")
            for bit_section in bit_sections_list:
                self.add_bit_section(bit_section=bit_section)
            return
        
        if bit_section is not None:
            if len(bit_section) < 3 or len(bit_section) > 5:
                raise TechnicalException(f"bit_section must be an iterable of length between 3 and 5")
            
            # Declare bit section
            bs = NamedTuple('BitSection', name=str, length=int, type=type, value=int, right_padded=bool)
            
            if isinstance(bit_section[0], str):
                bs.name = bit_section[0]
            else:
                raise TechnicalException(f"First bit_section element is the name and must be a string (obtained type: {Typing.get_object_class_fullname(bit_section[0])})")
            
            if isinstance(bit_section[1], int) and bit_section[1] >= 0:
                bs.length = bit_section[1]
            else:
                raise TechnicalException(f"Second bit_section element is the bit length and must be a positive integer (obtained type: {Typing.get_object_class_fullname(bit_section[1])} ; value: {bit_section[1]})")
            
            if isinstance(bit_section[2], type):
                bs.type = bit_section[2]
            else:
                raise TechnicalException(f"Third bit_section element is the value type and must be a type in (int, str, bytes) (obtained type: {Typing.get_object_class_fullname(bit_section[2])} ; value: {bit_section[2]})")
            
            # Set right padding
            if len(bit_section) > 4:
                if bit_section[4] is None or isinstance(bit_section[4], bool):
                    bs.right_padded = bit_section[4]
                else:
                    raise TechnicalException(f"Fifth bit_section element define if right padding must be done and must be a boolean or None (obtained type: {Typing.get_object_class_fullname(bit_section[4])} ; value: {bit_section[4]})")
            else:
                bs.right_padded = None
                
            self.__series.append(bs)
            self.__len = None
            self.__update_index_by_name()
            
            # Set value
            if len(bit_section) > 3:
                self.set_bit_section_value(bs.name, bit_section[3])
    
    def __update_index_by_name(self):
        self.__index_by_name = {}
        for index, bs in enumerate(self):
            self.__index_by_name[bs.name] = index
            
    def get_bit_section(self, index=None, name=None):
        if name is not None:
            if name in self.__index_by_name:
                index = self.__index_by_name[name]
            else:
                raise FunctionalException(f"No bit section of name '{name}'")

        if index is not None:
            return self.__series[index]
        else:
            raise TechnicalException("Undefined parameter 'index' or 'name'")
    
    def set_bit_section_value(self, name=None, value=None, index=None):
        bs = self.get_bit_section(index=index, name=name)
        if value is None:
            bs.value = value
        elif bs.type is int:
            if not isinstance(value, int):
                raise FunctionalException(f"Unexpected value type {type(value)} (bit section '{bs.name}' has type {bs.type})")
            bs.value = value
        elif bs.type is str:
            if not isinstance(value, str):
                raise FunctionalException(f"Unexpected value type {type(value)} (bit section '{bs.name}' has type {bs.type})")
            if StrTools.is_hex(value):
                bs.value = value
            else:
                raise TechnicalException(f"If value is set in string format, it has to be an hexa string (value: [{value}])")
        elif bs.type is bytes:
            if not isinstance(value, bytes):
                raise FunctionalException(f"Unexpected value type {type(value)} (bit section '{bs.name}' has type {bs.type})")
            bs.value = value
        else:
            raise TechnicalException(f"Value must be a positive integer, a hexa string, bytes or None (value type: {Typing.get_object_class_fullname(value)} ; value: {value})")
    
    def from_hex(self, hex_str, right_padded=False):
        """
        Fill bit series from given hexadecimal string.
        
        If bit series has a bit length that is not a multiple of 4, padding is expected in hex_str at left or right depending on parameter right_padded.
        A control is made if padding bits are only zeros.
        """
        # Verify padding
        bin_str = Binary.convert_hex_str_to_bin_str(hex_str)
        nb_bits_out_of_block = len(bin_str) - self.nb_bits
        if nb_bits_out_of_block < 0:
            raise FunctionalException(f"Hexadecimal string has not enough bits (expected: {self.nb_bits} ; obtained: {len(hex_str) * 4}")
        elif nb_bits_out_of_block > 0:
            if right_padded:
                padding_bits = bin_str[-nb_bits_out_of_block:]
            else:
                padding_bits = bin_str[0:nb_bits_out_of_block]
            value = int(padding_bits, 2)
            if value != 0:
                raise FunctionalException(f"Hexadecimal string has unexpected padded characters, only zero padding is managed (expected bits number: {self.nb_bits} ; {'right' if right_padded else 'left'} padding ; obtained padding bits: {padding_bits})\n    obtained bits: {bin_str}")
        
        # Fill bit sections
        bit_offset = 0 if right_padded else nb_bits_out_of_block
        for bs in self:
            # Extract bit section value
            try:
                hex_value = self.__from_hex_get_bit_section_value(hex_str, bit_offset=bit_offset, bit_length=bs.length, right_padding=bs.right_padded)
                if bs.type is int:
                    value = int(hex_value, 16)
                elif bs.type is bytes:
                    value = StrTools.hex_to_bytes(hex_value)
                elif bs.type is str:
                    value = hex_value
                else:
                    raise TechnicalException(f"Unmanaged bit section value type {bs.type} (possible types: int, str, bytes)")
            except Exception as exc:
                raise TechnicalException(f"Error while extracting value of field '{bs.name}': {exc}") from exc
                
            # Set value
            self.set_bit_section_value(bs.name, value)
            
            # Prepare next round
            bit_offset += bs.length
    
    def __from_hex_get_bit_section_value(self, hex_str, bit_offset, bit_length, right_padding):
        bin_str = "".join(["{:04b}".format(int(hc, 16)) for hc in hex_str])
        bs_bin_str = bin_str[bit_offset : bit_offset + bit_length]
        if bit_length % 4 > 0:
            str_padding = "0" * (4 - bit_length % 4)
            if right_padding is not None and right_padding:
                bs_bin_str_padded = bs_bin_str + str_padding
            else:
                bs_bin_str_padded = str_padding + bs_bin_str
        else:
            bs_bin_str_padded = bs_bin_str
        bs_hex_str = "".join(["{:X}".format(int(bs_bin_str_padded[i*4:i*4+4],2)) for i in range(len(bs_bin_str_padded)//4)])
        # logger.info(f"++++++++++ __from_hex_get_bit_section_value: bit_offset={bit_offset} ; bit_length={bit_length}\n    hex_str: [{hex_str}]\n    bin_str: [{bin_str}]\n    bs_bin_str: [{bs_bin_str}]\n    bs_bin_str: [{bs_bin_str_padded}]\n => bs_hex_str: [{bs_hex_str}]")
        return bs_hex_str
    
    def to_bin(self, right_padding=False):
        """
        Convert bit series to binary string.
        
        If a bit section is of type str and has a bit length lower than its section value,
        the section value is considered padded with zeros at left or right, depending on parameter right_padding.
        """
        res_list = []
        for bs in self:
            try:
                if bs.right_padded is not None:
                    bs_right_padded = bs.right_padded
                else:
                    bs_right_padded = right_padding
                    
                if bs.type is int:
                    bs_format = f"{{:0{bs.length}b}}"
                    res_bs = bs_format.format(bs.value)
                elif bs.type is str or bs.type is bytes:
                    if bs.type is bytes:
                        value = StrTools.to_hex(bs.value)
                    else:
                        value = bs.value
                    
                    res_bs = "".join(["{:04b}".format(int(hc, 16)) for hc in value])
                    if len(res_bs) > bs.length:
                        # The section value was padded
                        if bs_right_padded:
                            res_bs_padding = res_bs[bs.length:]
                            res_bs = res_bs[0:bs.length]
                        else:
                            res_bs_padding = res_bs[0:-bs.length]
                            res_bs = res_bs[-bs.length:]
                        if int(res_bs_padding, 2) != 0:
                            raise TechnicalException(f"For field '{bs.name}', the value was padded with {len(res_bs_padding)} bits ('{res_bs_padding}') at {'right' if right_padding else 'left'} with non zero bits")
                else:
                    raise TechnicalException(f"For field '{bs.name}', unmanaged type {bs.type}")
            except TechnicalException as exc:
                raise exc
            except Exception as exc:
                raise TechnicalException(f"Error while formatting field '{bs.name}': {exc}") from exc
            
            if len(res_bs) != bs.length:
                raise FunctionalException(f"For field '{bs.name}', the value [{bs.value}] has binary length {len(res_bs)} (expected length: {bs.length})")
            res_list.append(res_bs)
        return "".join(res_list)
    
    def to_hex(self, right_padding=False, bytes_padding=True, nbbits_by_block=None):
        """
        Convert bit series to hexadecimal string (ex: 'A2FF').
        
        If bit series has a length that is not a multiple of block size, a padding with zeros is made at left or right, depending on parameter right_padding.
        The block size is equal to:
            1. nbbits_by_block (if nbbits_by_block is not None)
            2. 8 (else if bytes_padding is True)
            3. 4 (in other cases)
        """
        # Convert to binary string
        res_bin = self.to_bin(right_padding=right_padding)

        # Convert binary string to hexadecimal string
        if nbbits_by_block is None:
            nbbits_by_block = 8 if bytes_padding else 4
        return Binary.convert_bin_str_to_hex_str(res_bin, right_padding, nbbits_by_block=nbbits_by_block)
    
def convert_bit_series_to_table(bit_series):
    res = TableWithHeader()
    res.header = TableRow(cells_content=["Name", "Bit length", "Type", "Value"])
    for bit_section in bit_series:
        res.add_row(cells_content=[bit_section.name, bit_section.length, bit_section.type, bit_section.value])
    return res

def convert_bit_series_to_name_value_table(bit_series):
    res = TableWithHeader()
    res.header = TableRow(cells_content=["Name", "Value"])
    for bit_section in bit_series:
        res.add_row(cells_content=[bit_section.name, bit_section.value])
    return res





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
import copy
from holado_helper.script.input_output_script import InputOutputScript
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_python.standard_library.typing import Typing
from holado_helper.script.action import ActionInfo
from holado_core.common.actors.actor import Actor
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.common.tools.datetime import DateTime
from holado.common.handlers.object import Object
from holado.common.handlers.undefined import undefined_argument
from holado_helper.script.job import Job

logger = logging.getLogger(__name__)


class ActionScriptFile():
    def __init__(self, file_ext):
        super().__init__()
        self.__file_ext = file_ext
    
    @property
    def file_extension(self):
        return self.__file_ext
    
    def read_jobs(self, file_path, action_script):
        pass
    
    def write_jobs(self, file_path, jobs, action_script):
        pass
    


class ActionRegistry(Object):
    def __init__(self, name):
        super().__init__(name)
        
        self.__actions = {}
    
    def register_action(self, action, parameters_default_values=None, parameters_replacements=None, static_parameters_values=None, **kwargs):
        """ Register new action
        @param action: Action name
        @param parameters_default_values: define parameters default values (format: list of 2-uplets (parameter name, default value))
        @param parameters_replacements: define values automatic replacements (format: dictionary of {parameter name: List of 2-uplets (alias value, replacement value)})
        @param static_parameters_values: fixed parameter values known at action registration (format: dictonary of {parameter name: parameter value})
        @param kwargs: any other action arguments
        """
        action_kwargs = {
            'parameters_default_values': parameters_default_values, 
            'parameters_replacements': parameters_replacements,
            'static_parameters_values': static_parameters_values
            }
        action_kwargs.update(kwargs)
        self.__actions[action] = action_kwargs
        logger.info(f"[{self.name}] Registered action '{action}'")
    
    def get_action_names(self):
        return list(self.__actions.keys())
    
    def get_action_parameter_names(self, action):
        parameters_default_values = self.get_action_kwargs(action)['parameters_default_values']
        return [p[0] for p in parameters_default_values]
    
    def get_action_parameters_default_values(self, action):
        return self.get_action_kwargs(action)['parameters_default_values']
    
    def get_action_parameters_replacements(self, action):
        return self.get_action_kwargs(action)['parameters_replacements']
    
    def get_action_static_parameters_values(self, action):
        return self.get_action_kwargs(action)['static_parameters_values']
    
    def get_action_kwargs(self, action):
        if action in self.__actions:
            return self.__actions[action]
        else:
            raise TechnicalException(f"No registered action '{action}'")
    



class ActionScriptActor(Actor):
    def __init__(self, module_name, action_registry=None):
        super().__init__(module_name)
        self.__action_registry = action_registry
    
    @property
    def action_registry(self):
        return self.__action_registry
    
    @action_registry.setter
    def action_registry(self, action_registry):
        self.__action_registry = action_registry
    
    def act_process_action(self, action_info, action_script):
        raise NotImplementedError()


class ActionScript(InputOutputScript):
    def __init__(self, work_dir_path=None, with_input=True, with_output=True, with_template=True, with_action_parameter=True, 
                 actor=None, action_registry=None, script_file=None):
        super().__init__(work_dir_path=work_dir_path, with_input=with_input, with_output=with_output, with_template=with_template)
        
        self.__actor = actor
        self.__action_registry = action_registry if action_registry is not None else ActionRegistry("default")
        self.__script_file = script_file
        self.__with_action_parameter = with_action_parameter
        
        # Post process
        if self.__actor and self.__actor.action_registry is None:
            self.__actor.action_registry = self.__action_registry
    
    @property
    def actor(self):
        return self.__actor
    
    @property
    def action_registry(self):
        return self.__action_registry
    
    @property
    def script_file(self):
        return self.__script_file
    
    @property
    def with_action_parameter(self):
        return self.__with_action_parameter
    
    def get_description(self, script_command, short_description, action):
        return """{short_description}
        
        Usage:
            1. Generate a template file for the desired action with parameter '--output-template' as in example 1.
                It is possible to specify the name of the output file with parameter '-o'.
            2. Copy the template line in order to have one line per action to do, and modify the template file with desired values. 
            3. Process actions described in a CSV file with parameter '-i'.
                If an error occurs at a line, execution is stopped.
            4. See result in output file. If parameter '-o' is not specified, the output filename is generated from parameter '-i'.
        
        In template files, a default value is set in each column. 
        For columns with enum values, the default value is a combination of each possible enum values separated by a '|', but only one value is possible.
        
        
        In input and output files, the first column 'STATUS' describes the status of execution of each line:
            - 'TODO' (or ''): the line is to process (in output file, it means that execution was stopped on an error with a previous line)
            - 'SUCCESS': the line was in success
            - 'ERROR': the line was in error
            - others: the line is not processed, you can specify your own status to skip this line
        In template files, this column is automatically initialized with status 'TODO'.
        Any output file can be used as input file, the lines that are in status 'ERROR' or 'TODO' will be processed again.
        
        Example 1: Output CSV template for action '{action}'
                
                {script_command} {action} --output-template -o {action}.csv
                
        Example 2: Process action '{action}' with given input CSV
                
                {script_command} {action} -i {action}.csv
        """.format(short_description=short_description, script_command=script_command, action=action)
        
    def _new_argument_parser(self, description=None, additional_arguments=None, add_help=True):
        parser = super()._new_argument_parser(description=description, additional_arguments=additional_arguments, add_help=add_help)
        
        if self.__with_action_parameter:
            parser.add_argument('action', choices=self.possible_actions(), default=None, help='Action name')
        
        return parser
    
    def is_matching_file_path(self, file_path, raise_exception=False):
        file_ext = self.script_file.file_extension.lower()
        res = file_path.lower().endswith(file_ext)
        
        if not res and raise_exception:
            raise VerifyException(f"File path doesn't end by {file_ext} (case insensitive)")
        return res
    
    def is_matching_action(self, action, raise_exception=False):
        """Return if action is matching this action script
        """
        res = action in self.possible_actions()
        
        if not res and raise_exception:
            raise VerifyException(f"Action '{action}' doesn't match this script (possible actions: {self.possible_actions()})")
        return res
    
    def is_matching_action_and_input_file_path(self, action, file_path, raise_exception=False):
        """Return if action and associated input file are matching this action script
        """
        return self.is_matching_action(action, raise_exception) \
            and self.is_matching_input_file_path(file_path, raise_exception=raise_exception)
    
    def is_matching_action_and_output_file_path(self, action, file_path, raise_exception=False):
        """Return if action and associated output file are matching this action script
        """
        return self.is_matching_action(action, raise_exception) \
            and self.is_matching_output_file_path(file_path, raise_exception=raise_exception)
    
    def _build_output_template_path(self):
        if self.with_action_parameter:
            default_path = f"{self.args.action}_template{self.script_file.file_extension}"
        else:
            default_path = None
        
        try:
            return self._build_output_path(self.args.output_filename, default_path, self.user_working_dir)
        except Exception:
            raise FunctionalException(f"[{Typing.get_object_class_fullname(self)}] Please define output template file path")
    
    def possible_actions(self):
        raise NotImplementedError
    
    def build_template_jobs(self):
        if self.with_action_parameter:
            action = self.args.action
            params_default_values = self.action_registry.get_action_parameters_default_values(action)
            params = {pdv[0]:pdv[1] for pdv in params_default_values}
            # Note: for templates, add static parameters is not needed
            actions = [self.new_action_info(action, params=params)]
            job = Job(actions=actions, in_parallel=False)
            return [job]
        else:
            raise NotImplementedError()
    
    def process_jobs(self, list_jobs):
        res = []
        got_error = False
        
        for job in list_jobs:
            if got_error:
                res.append(job)
                continue
            
            res_job, got_error = self.process_job(job)
            res.append(res_job)
            
        return res, got_error
    
    def process_job(self, job):
        return self.process_actions(job.actions, job.in_parallel)
    
    def new_action_info_from_args(self):
        action = self.args.action
        static_params = self.action_registry.get_action_static_parameters_values(action)
        return self.new_action_info(action, static_params=static_params)
    
    def new_action_info(self, action, uid=None, params=None, static_params=None, index=0):
        processing_params = self.build_processing_action_params(action, action_args=self.args if self.with_action_parameter else None, action_params=params)
        if uid is None:
            dt_str = DateTime.datetime_2_str(DateTime.utcnow())
            uid = f"{dt_str} - {action} - {processing_params}"
        return ActionInfo(uid=uid, action=action, params=processing_params, static_params=static_params, index=index)
    
    def build_processing_action_params(self, action, action_args=None, action_params=None, **kwargs):  # @UnusedVariable
        if action_params is not None:
            return dict(action_params)
        else:
            return None
    
    def process_actions(self, list_action_info, in_parallel):
        if in_parallel:
            return self._process_actions_parallel(list_action_info)
        else:
            return self._process_actions_serial(list_action_info)
        
    def _process_actions_serial(self, list_action_info):
        logger.info(f"Processing {len(list_action_info)} actions in serial mode")
        res = []
        got_error = False
        
        for action_info in list_action_info:
            action_status = action_info.params['STATUS'] if 'STATUS' in action_info.params else 'TODO'
            res_action = copy.deepcopy(action_info)
            
            if not got_error and action_status in ['TODO', 'ERROR']:
                action_status = self.process_action(action_info)
                got_error |= (action_status == 'ERROR')
            elif got_error:
                logger.info(f"Skip action of index {action_info.index} since a previous action has failed")
            else:
                logger.info(f"Skip action of index {action_info.index} since its statud is '{action_status}'")
                
            res_action.params['STATUS'] = action_status
            res.append(res_action)
                
        return res, got_error
        
    def _process_actions_parallel(self, list_action_info):
        from holado_multitask.multiprocessing.function_process import FunctionProcess
        
        logger.info(f"Processing {len(list_action_info)} actions in parallel mode")
        res = []
        got_error = False
        processes = []
        
        # Start all processes
        for action_info in list_action_info:
            action_status = action_info.params['STATUS'] if 'STATUS' in action_info.params else 'TODO'
            res_action = copy.deepcopy(action_info)
            
            if action_status in ['TODO', 'ERROR']:
                proc = FunctionProcess(self.process_action, args=[action_info], name=f"Process action of index {action_info.index}")
                proc.start()
                processes.append((action_info, proc))
            
            res_action.params['STATUS'] = action_status
            res.append(res_action)
            
        # Join all processes
        for _, proc in processes:
            proc.join()
            
        # Update result
        for action_info, proc in processes:
            if proc.error is not None:
                logger.error(f"Error when processing action of index {action_info.index}: {proc.error}")
                action_status = 'ERROR'
            if proc.result is not None:
                action_status = proc.result
            
            res[action_info.index].params['STATUS'] = action_status
            got_error |= (action_status == 'ERROR')
            
        return res, got_error
        
    def _apply_replacements_on_action_params(self, action_params, params_replacements):
        from holado_core.common.tools.tools import Tools
        
        res = dict(action_params)
        
        for key in action_params:
            value = action_params[key]
            if key in params_replacements:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Checking replacable value '{key}'")
                for value_replaced, replacement in params_replacements[key]:
                    value_comp = value.lower() if isinstance(value, str) else value
                    value_replaced_comp = value_replaced.lower() if isinstance(value_replaced, str) else value_replaced
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"Evaluating value '{value_comp}' to '{value_replaced_comp}'. Value for '{key}' replaced by '{replacement}' if equals")
                    if value_comp == value_replaced_comp:
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"Value for '{key}' replaced by '{replacement}'")
                        value = replacement
                        break
                res[key] = value
        
        return res
        
    def process_action(self, action_info):
        if self.actor:
            return self.actor.act_process_action(action_info, self).execute()
        else:
            raise NotImplementedError()
    
    def read_jobs(self, input_path):
        res = self.script_file.read_jobs(input_path, self)
        
        # Verify action parameter names
        self._verify_jobs_actions_parameters(res)
        
        # Replace action parameters if needed
        self._replace_jobs_actions_parameters(res)
        
        return res
    
    def _verify_jobs_actions_parameters(self, jobs):
        for job in jobs:
            for action_info in job.actions:
                self._verify_action_parameters(action_info)
    
    def _verify_action_parameters(self, action_info):
        ai_param_names = set(action_info.params.keys())
        expected_param_names = set(self.action_registry.get_action_parameter_names(action_info.action))
        difference = ai_param_names.symmetric_difference(expected_param_names)
        # Note: parameter 'STATUS' is added in action result and is possibly present in input
        if len(difference) > 0 and difference != {'STATUS'}:
            raise FunctionalException(
                f"""Unexpected action parameter names for action '{action_info.action}':
                    obtained names: {ai_param_names}
                    expected names: {expected_param_names}""")

    def _replace_jobs_actions_parameters(self, jobs):
        for job in jobs:
            for action_info in job.actions:
                self._replace_action_parameters(action_info)
    
    def _replace_action_parameters(self, action_info):
        from holado_core.common.tools.tools import Tools
        
        # Process replacements defined in action registry
        params_replacements = self.action_registry.get_action_parameters_replacements(action_info.action)
        if params_replacements:
            for param_name in action_info.params:
                value = action_info.params[param_name]
                if param_name in params_replacements:
                    new_value = undefined_argument
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"Checking replacable value '{param_name}'")
                    for value_replaced, replacement in params_replacements[param_name]:
                        value_comp = value.lower() if isinstance(value, str) else value
                        value_replaced_comp = value_replaced.lower() if isinstance(value_replaced, str) else value_replaced
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.trace(f"Comparing value '{value_comp}' to '{value_replaced_comp}'. Value for parameter '{param_name}' replaced by '{replacement}' if equals")
                        if value_comp == value_replaced_comp:
                            new_value = replacement
                            break
                    if new_value is not undefined_argument:
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"For parameter '{param_name}', value '{value}' is replaced by '{new_value}'")
                        action_info.params[param_name] = new_value
    
    def write_jobs(self, output_path, result_jobs, got_error):
        output_path_base, _, output_ext = self._build_output_path_base(output_path)
        if got_error is not None:
            if got_error:
                output_base = f"{output_path_base}-FAIL"
            else:
                output_base = f"{output_path_base}-SUCCESS"
        else:
            output_base = output_path_base
        real_output_path = self._build_output_path_from_base(output_base, output_ext)
        
        self.script_file.write_jobs(real_output_path, result_jobs, self)
        
        return real_output_path
    
    def run_script(self):
        if self.with_template and self.args.output_template:
            output_path = self._build_output_template_path()
            template_jobs = self.build_template_jobs()
            real_output_path = self.write_jobs(output_path, template_jobs, got_error=None)
            logger.print(f"Template created in file '{real_output_path}'")
        elif self.with_input and self.args.input_filename:
            # Read file
            input_path = self._build_input_path(self.args.input_filename, self.user_working_dir)
            list_jobs = self.read_jobs(input_path)
            
            # Process jobs
            result_jobs, got_error = self.process_jobs(list_jobs)
            
            # Write file
            if self.with_output and self.args.output_filename is not None:
                output_path = self._build_output_path(self.args.output_filename, self.args.input_filename, self.user_working_dir)
                real_output_path = self.write_jobs(output_path, result_jobs, got_error)
                logger.print(f"Result outputed in file '{real_output_path}'")
        else:
            action_info = self.new_action_info_from_args()
            result_action = self.process_action(action_info)
            result_jobs=[{'actions': [result_action]}]
            
            # Write file
            if self.with_output and self.args.output_filename is not None:
                output_path = self._build_output_path(self.args.output_filename, None, self.user_working_dir)
                real_output_path = self.write_jobs(output_path, result_jobs)
                logger.print(f"Result outputed in file '{real_output_path}'")






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_ui.ui.drivers.ui_driver import UIDriver
from queue import LifoQueue
from holado_core.common.exceptions.technical_exception import TechnicalException
import re
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class GUIDriver(UIDriver):
    """ Base class for GUI drivers.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name=None):
        super().__init__(name or "GUIDriver")
        
        # Information on opened pages
        self.__window_id_by_visible_order = LifoQueue()
        self.__window_info_by_id = {}
        
        # Settings used to implement the right GUIWindow instance when a new window is opened
        self.__window_type_by_name_pattern = {}
        self.__default_window_type = None
    
    def _close_driver(self):
        # Close each opened page before 
        while not self.__window_id_by_visible_order.empty():
            self._close_current_window()
    
    def _close_current_window(self):
        raise NotImplementedError

    def _has_window_info(self, window_id):
        return window_id in self.__window_info_by_id
    
    def _add_window_info(self, window_id, window_info):
        self.__window_info_by_id[window_id] = window_info
    
    def _get_window_info(self, window_id):
        res = self.__window_info_by_id[window_id]
        if res is None:
            msg = f"No descriptor exists for ID '{window_id}'. Currently known IDs: {list(self.__window_info_by_id.keys())}"
            raise TechnicalException(msg)
        return res
    
    def _update_window_info(self, window_id, window):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Updated window info: ID={window_id} ; window type={Typing.get_object_class_fullname(window)}")
        window_info = self._get_window_info(window_id)
        window_info.window = window
    
    def _clear_window_info(self):
        self.__window_id_by_visible_order.clear()
        self.__window_info_by_id.clear()
    
    @property
    def default_window_type(self):
        return self.__default_window_type
    
    @default_window_type.setter
    def default_window_type(self, default_type):
        """
        Set the type to use by default when a new window is opened.
        @param default_type Default window type
        """
        self.__default_window_type = default_type
    
    def _get_window_type(self, window_name):
        res = None
        
        # Get window type by name
        for name_pattern in self.__window_type_by_name_pattern:
            m = re.compile(name_pattern).match(window_name)
            if m:
                res = self.__window_type_by_name_pattern[name_pattern]
                break
        
        # Else get default window type
        if res is None:
            res = self.default_window_type
        
        return res

    def add_window_type(self, window_name_pattern, window_type):
        """
        Set the type to use for given window name pattern.
        Note: window name patterns are ordered  thus, if two patterns match a window name, the first added will be used
        @param window_namePattern Pattern of window name
        @param window_type Window type
        """
        self.__window_type_by_name_pattern[window_name_pattern] = window_type
    
    def _has_current_window(self):
        return not self.__window_id_by_visible_order.empty()
    
    def _nb_windows(self):
        return self.__window_id_by_visible_order.qsize()
    
    def _get_all_window_ids(self):
        return list(self.__window_id_by_visible_order.queue)
    
    def _get_current_window_id(self):
        if self.__window_id_by_visible_order.empty():
            raise FunctionalException("No window is currently opened.")
        return self.__window_id_by_visible_order.queue[-1]
    
    def _pop_current_window_id(self):
        if self.__window_id_by_visible_order.empty():
            raise FunctionalException("No window is currently opened.")
        return self.__window_id_by_visible_order.get()
    
    def _push_current_window_id(self, window_id):
        if self._has_window_info(window_id):
            logger.warning(f"Repush window {window_id} on top")
        self.__window_id_by_visible_order.put(window_id)
    
    @property
    def current_window(self):
        """
        @return current window.
        """
        return self._get_window_info(self._get_current_window_id()).window
    
    def update_current_window(self, window_name, window_id):
        """
        Update current window instance after a new window has been opened
        @param window_name Window name
        @param window_id Window ID (for verification)
        """
        # Verify that window ID is current window
        if self._get_current_window_id() != window_id:
            raise FunctionalException("A new window was opened.")
        
        # Find type specific to displayed window
        window = self.new_window_for_name(window_name)
        self._update_window_info(self.get_current_window_id(), window)
    
    def update_current_window_with_default(self, window_id):
        """
        Update current window instance after a new window has been opened, with default window
        @param window_id Window ID (for verification)
        """
        # Verify that window ID is current window
        if self.get_current_window_id() != window_id:
            raise FunctionalException("A new window was opened.")

        # Find type specific to displayed window
        window = self.new_default_window()
        self._update_window_info(self.get_current_window_id(), window)
    
    def new_default_window(self):
        """
        @return New instance of default window
        """
        specific_type = self.default_window_type
        if specific_type is not None:
            return self.__new_window_for_type(specific_type)
        else:
            raise TechnicalException("Unable to determine default window type")
        
    def _new_window_for_name(self, window_name):
        specific_type = self._get_window_type(window_name)
        if specific_type is not None:
            return self.__new_window_for_type(specific_type)
        else:
            raise TechnicalException(f"Unable to determine window type for window name '{window_name}'")
        
    def __new_window_for_type(self, specific_type):
        if specific_type is None:
            raise TechnicalException("Type is None")

        res = specific_type()
        res.initialize(self)

        return res
    
    def make_screenshots_for_debug(self, destination_path, context_description):
        """
        Make screenshots for debug.
        @param destination_path Destination path
        @param context_description Context description that will be inserted in file names
        """
        raise NotImplementedError
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc

logger = logging.getLogger(__name__)



class BaseAction(object):
    """ Base class for actions.
    An action class defines how to process an action, and then it can be processed with different argument values.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name, func_execute):
        self.__name = name
        self.__func_execute = func_execute
        
    @property
    def name(self):
        return self.__name
    
    @property
    def function_execute(self):
        return self.__func_execute
    
    @function_execute.setter
    def function_execute(self, func):
        self.__func_execute = func
    
    def execute(self, *args, **kwargs):
        """ Execute action
        @return Result or None
        """
        raise NotImplementedError
        
    
class Action(BaseAction):
    """ Action without input.
    """
    def __init__(self, name, func_execute):
        super().__init__(name, func_execute)
        
    def execute(self, *args, **kwargs):
        """ Execute action
        @return Result or None
        """
        return self.function_execute(*args, **kwargs)
    
    
class FindAction(BaseAction):
    """ Action launching Finder.find method with given find type.
    """
    def __init__(self, name, find_builder, func_execute=None, finder=None, find_type=None):
        super().__init__(f"find '{finder.element_description}'" if name is None and finder is not None else name, func_execute)
        self.__find_builder = find_builder
        
        if finder is not None:
            def f(*args, **kwargs):
                return finder.find(find_type, *args, **kwargs)
            self.function_execute = f
        
    def execute(self, container=None, candidates=None, find_context=None, find_parameters=None):
        container, candidates, find_context, find_parameters = self.__find_builder.update_find_inputs(container, candidates, find_context, find_parameters)
        return self.function_execute(container, candidates, find_context, find_parameters)
    
class FindAllAction(BaseAction):
    """ Action launching Finder.find_all method with given find type.
    """
    def __init__(self, name, find_builder, func_execute=None, finder=None, find_type=None):
        super().__init__(f"find all '{finder.element_description}'" if name is None and finder is not None else name, func_execute)
        self.__find_builder = find_builder
        
        if finder is not None:
            def f(*args, **kwargs):
                return finder.find_all(find_type, *args, **kwargs)
            self.function_execute = f
        
    def execute(self, container=None, candidates=None, find_context=None, find_parameters=None):
        container, candidates, find_context, find_parameters = self.__find_builder.update_find_inputs(container, candidates, find_context, find_parameters)
        return self.function_execute(container, candidates, find_context, find_parameters)
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import object
import logging
from holado.common.handlers.enums import ObjectStates

logger = None

def initialize_logger():
    global logger
    logger = logging.getLogger(__name__)


class Object(object):
    def __init__(self, name=None):
        self.__name = name
        self.__obj_state = ObjectStates.Created

    @property
    def name(self):
        if self.__name:
            return self.__name
        else:
            return str(self)

    @property
    def object_state(self):
        return self.__obj_state
    
    @object_state.setter
    def object_state(self, state):
        from holado_core.common.tools.tools import Tools
        from holado_python.standard_library.typing import Typing
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"State changed for object '{self.name}' (type: {Typing.get_object_class_fullname(self)}): {self.__obj_state} -> {state}")
        self.__obj_state = state


class DeleteableObject(Object):
    """
    Add to a class deleting features:the possibility to be deleted
        - Self deleting actions
        - Register functions to call on delete
    
    Note: It is highly recommended to set garbage collector periodicity with GcManager.collect_periodically method (see method documentation)
    """
    
    def __init__(self, name=None):
        super().__init__(name)
        
        self.__on_delete_funcs = []
        self.__on_delete_gc_collect = False
        self.__is_deleteable = True
    
    @property
    def _is_deleteable(self):
        return self.__is_deleteable
    
    @_is_deleteable.setter
    def _is_deleteable(self, is_deleteable):
        self.__is_deleteable = is_deleteable
    
    def __del__(self):
        try:
            if not self._is_deleteable:
                return
            if self.object_state in [ObjectStates.Deleting, ObjectStates.Deleted]:
                return
            
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Deleting object of ID {id(self)} and type {type(self)}")
            if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                import traceback
                logger.trace("".join(traceback.format_list(traceback.extract_stack())))
            
            # Notes: 
            #   - Garbage collector default mecanism is incompatible with custom object finalizers (cf GcManager.Method __del__ is called by garbage collector. This call can be done in any thread at any moment.
            #   - When SessionContext is under deletion, delete any object by thread is not possible
            from holado.common.context.session_context import SessionContext
            from holado.common.tools.gc_manager import GcManager
            if GcManager.is_collect_threaded() or isinstance(self, SessionContext.TSessionContext) or not SessionContext.has_instance():
                self.__delete()
            else:
                from holado_multitask.multithreading.functionthreaded import FunctionThreaded
                func = FunctionThreaded(self.__delete, name=f"delete object '{self.name}'", register_thread=False, do_log=False)
                func._is_deleteable = False      # Deactivate deletion mechanism of DeleteableObject to avoid an infinite loop
                func.start()
                func.join(timeout=30, raise_if_still_alive=False)   # timeout of 30s to limit deadlock impact ; raise to False since exceptions are omitted in gc collect context
        except AttributeError as exc:
            if "'NoneType' object has no attribute" in str(exc):
                # Simply return, this exception can appear when system in interrupting
                return
            else:
                raise exc
        except Exception as exc:
            if "Python is likely shutting down" in str(exc):
                # Simply return
                return
            else:
                raise exc
    
    def __delete(self):
        try:
            self.delete_object()
        except Exception as exc:
            if "Python is likely shutting down" in str(exc):
                # Simply return
                return
            else:
                raise exc
        
    @property
    def on_delete_call_gc_collect(self):
        return self.__on_delete_gc_collect
        
    @on_delete_call_gc_collect.setter
    def on_delete_call_gc_collect(self, do_call):
        self.__on_delete_gc_collect = do_call
        
    def on_delete_call_func(self, func):
        self.__on_delete_funcs.append(func)
        
    def delete_object(self):
        from holado_core.common.tools.tools import Tools
        from holado_python.standard_library.typing import Typing
        
        if not self._is_deleteable:
            return
        if self.object_state in [ObjectStates.Deleting, ObjectStates.Deleted]:
            return False
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Deleting object '{self.name}' (type: {Typing.get_object_class_fullname(self)})...")
        self.object_state = ObjectStates.Deleting
        
        # Run on_delete functions
        for func in self.__on_delete_funcs:
            try:
                func()
            except:
                logger.exception(f"[{self.name}] Error while running on delete function '{func}'")
                
        self._delete_object()
        
        # Call garbage collector
        if self.__on_delete_gc_collect:
            from holado.common.tools.gc_manager import GcManager
            GcManager.collect()
        
        self.object_state = ObjectStates.Deleted
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Deleted object '{self.name}' (type: {Typing.get_object_class_fullname(self)})")
        
        return True

    def _delete_object(self):
        """
        Implement this method for specific delete actions
        """
        pass
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import threading
import time
import logging
import holado_multitask.multithreading.thread as ctt
from holado.common.handlers.undefined import default_context
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)


class LoopThread(ctt.InterruptableThread):
    """ Helper class managing a loop in a thread."""

    def __init__(self, name, default_wait_timeout=None, register_thread=True, register_scope=default_context, delay_before_run_sec=None, delay_between_run_sec=None, nb_times=None, with_next_loop_datetime=False):
        super().__init__(name, default_wait_timeout=default_wait_timeout, register_thread=register_thread, register_scope=register_scope, delay_before_run_sec=delay_before_run_sec, delay_between_run_sec=delay_between_run_sec)
        
        self.__nb_times = nb_times
        
        # Manage loops
        self.__with_next_loop_datetime = with_next_loop_datetime
        self.__is_under_loop = threading.Event()
        self.__next_loop_datetime = DateTime.now() if self.__with_next_loop_datetime else None
        self.__loop_counter = 0
        
    @property
    def next_loop_datetime(self):
        return self.__next_loop_datetime

    @next_loop_datetime.setter
    def next_loop_datetime(self, dt):
        self.__next_loop_datetime = dt

    @property
    def loop_counter(self):
        return self.__loop_counter
        
    def run(self):
        logger.info("Loop thread '{}' is launched.".format(self.name))
        
        self.wait_before_run()
        
        counter = 0
        while not self.is_interrupted:
#             logger.debug("[Thread '{}'] New run".format(self.name))
            try:
                if self.does_need_run_loop():
                    with self._is_lock:
                        self._is_idle.clear()
                        self.__is_under_loop.set()
                        self.__loop_counter += 1
                    
                    self.run_loop()
                    counter += 1
                    
                    with self._is_lock:
                        self.__is_under_loop.clear()
                        self._is_idle.set()
                        
                    # Stop if nb_times is reached
                    if self.__nb_times is not None:
                        if counter >= self.__nb_times:
                            break
                    
                    if not self.__with_next_loop_datetime:
                        self.wait_between_run()
                else:
                    time.sleep(self._get_sleep_time_before_next_loop())
                    
            except:
                logger.exception(f"[Thread '{self.name}'] Exception catched during loop")

        logger.info(f"Loop thread '{self.name}' has finished.")
        
    def does_need_run_loop(self):
        if self.__with_next_loop_datetime:
            return DateTime.now() >= self.__next_loop_datetime
        else:
            return True
    
    def _get_sleep_time_before_next_loop(self):
        if self.__with_next_loop_datetime:
            now = DateTime.now()
            if now < self.__next_loop_datetime:
                delta = self.__next_loop_datetime - now
                res = delta.total_seconds()
                if res > 1:
                    return 1
                else:
                    return res
        return 0
    
    def run_loop(self):
        raise NotImplementedError()

    def wait_is_under_loop(self, timeout=None):
        self._wait(self.__is_under_loop, "is under loop", timeout=timeout)





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import os
import logging

logger = logging.getLogger(__name__)



class ResourceManager():
    """
    Manage local resources, ie resources stored on disk during and through sessions.
    For example, it can be used to persist data through sessions, so that a session can adapt its behavior according previous sessions.
    """
    def __init__(self, local_resource_path):
        self.__local_resource_path = local_resource_path
        
        self.__func_db_manager = None
        self.__func_path_manager = None
        self.__db_connect_kwargs_by_name = {}
        
    def initialize(self, func_db_manager, func_path_manager):
        self.__func_db_manager = func_db_manager
        self.__func_path_manager = func_path_manager
        
        self.__path_manager.makedirs(self.__local_resource_path, is_directory=True)
        
    @property
    def __db_manager(self):
        return self.__func_db_manager()
    
    @property
    def __path_manager(self):
        return self.__func_path_manager()
        
    @property
    def local_resource_path(self):
        return self.__local_resource_path
    
    def get_path(self, *args):
        return os.path.join(self.__local_resource_path, *args)
    
    def get_db_client(self, name, is_persistent=False):
        """ Return a SQLite3 DB client to the resource DB of given name
        @param is_persistent If True, name is prefixed with 'persistent/db/'
        """
        full_name = os.path.join("persistent", "db", name) if is_persistent else name
        _, res = self.__db_manager.get_or_create(full_name, 'sqlite3', self.__get_db_connect_kwargs(full_name))
        return res
    
    def __get_db_connect_kwargs(self, full_name):
        if full_name not in self.__db_connect_kwargs_by_name:
            db_filepath = self.get_path(f"{full_name}.sqlite3")
            self.__path_manager.makedirs(db_filepath)
            
            uri = f"file:{db_filepath}?mode=rwc"
            connect_kwargs = {'database': uri,
                              'uri': True}
            self.__db_connect_kwargs_by_name[full_name] = connect_kwargs
        return self.__db_connect_kwargs_by_name[full_name]
        
    def persist_pair(self, key, value, db_name="default", table_name="pair", do_commit=True):
        client = self.get_db_client(db_name, is_persistent=True)
        client.execute(f"create table if not exists {table_name} (key, value)", do_commit=do_commit)
        
        client.execute(f"insert into {table_name} values (?, ?)", key, value, do_commit=do_commit)
        
    def has_data_table(self, table_name, db_name="default", is_persistent=False):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        return client.exist_table(table_name)
        
    def create_data_table(self, table_name, create_sql, db_name="default", is_persistent=False, raise_if_exist=False, do_commit=True, do_audit=False):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        client.create_table(table_name, create_sql, raise_if_exist=raise_if_exist, do_commit=do_commit, do_audit=do_audit)
        
    def delete_data_table(self, table_name, db_name="default", is_persistent=False, raise_if_not_exist=False, do_commit=True):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        client.drop_table(table_name, raise_if_not_exist=raise_if_not_exist, do_commit=do_commit)
        
    def check_data_table_schema(self, table_name, create_sql, db_name="default", is_persistent=False):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        result = client.select("sqlite_schema", where_data={'name':table_name}, sql_return='sql')
        if not result:
            return False
        
        sql = result[0][0].content
        return sql == create_sql
        
    def count_data(self, table_name, where_data: dict=None, where_compare_data: list=None, db_name="default", is_persistent=False):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        return client.count(table_name, where_data=where_data, where_compare_data=where_compare_data)
        
    def has_data(self, table_name, where_data: dict=None, where_compare_data: list=None, db_name="default", is_persistent=False):
        count = self.count_data(table_name, where_data=where_data, where_compare_data=where_compare_data, db_name=db_name, is_persistent=is_persistent)
        return count > 0
        
    def get_data(self, table_name, where_data: dict=None, where_compare_data: list=None, db_name="default", is_persistent=False, result_as_dict_list=False, as_generator=False):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        result = client.select(table_name, where_data=where_data, where_compare_data=where_compare_data, result_as_dict_list=result_as_dict_list, as_generator=as_generator)
        return result
        
    def add_data(self, table_name, data: dict, db_name="default", is_persistent=False, do_commit=True):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        result = client.insert(table_name, data, do_commit=do_commit)
        return result
        
    def update_data(self, table_name, data: dict, where_data: dict=None, where_compare_data: list=None, db_name="default", is_persistent=False, do_commit=True):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        result = client.update(table_name, data=data, where_data=where_data, where_compare_data=where_compare_data, do_commit=do_commit)
        return result
        
    def delete_data(self, table_name, where_data: dict=None, where_compare_data: list=None, db_name="default", is_persistent=False, do_commit=True):
        client = self.get_db_client(db_name, is_persistent=is_persistent)
        client.delete(table_name, where_data=where_data, where_compare_data=where_compare_data, do_commit=do_commit)
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.handlers.undefined import default_value
from holado_db.tools.db.clients.base.db_audit import TriggerAuditManager

logger = logging.getLogger(__name__)


class SQLite3TriggerAuditManager(TriggerAuditManager):
    def __init__(self, name, db_client):
        super().__init__(name, db_client)
    
    def _get_audit_table_sql_create(self, audit_table_name=default_value):
        audit_table_name = self._get_audit_table_name(audit_table_name)
        return f"""CREATE TABLE {audit_table_name} (
                id INTEGER PRIMARY KEY,
                table_name TEXT,
                record_id INTEGER,
                operation_type TEXT,
                changed_at TEXT,
                previous_values TEXT,
                new_values TEXT
            )"""
    
    def _get_drop_trigger_sql_of_audit_table_operation(self, table_name, operation_type, audit_table_name):
        trigger_name = self.get_trigger_name(table_name, operation_type)
        return f"DROP TRIGGER IF EXISTS {trigger_name}"
    
    def _get_create_trigger_sql_to_audit_table_operation(self, table_name, operation_type, audit_table_name):
        audit_table_name = self._get_audit_table_name(audit_table_name)
        trigger_name = self.get_trigger_name(table_name, operation_type)
        trigger_insert_sql = self.__get_trigger_insert_sql(table_name, operation_type, audit_table_name)
        return f"""CREATE TRIGGER {trigger_name} AFTER {operation_type.upper()} ON {table_name}
                BEGIN
                    {trigger_insert_sql}
                END;
                """
    
    def __get_trigger_insert_sql(self, table_name, operation_type, audit_table_name):
        values = [f"'{table_name}'",
                  "NEW.rowid" if operation_type in ['insert', 'update'] else "OLD.rowid",
                  f"'{operation_type[0]}'",
                  "DATETIME('now')",
                  self.__get_row_version_json_object_sql(table_name, 'OLD') if operation_type in ['update', 'delete'] else "NULL",
                  self.__get_row_version_json_object_sql(table_name, 'NEW') if operation_type in ['insert', 'update'] else "NULL",
                 ]
        
        return f"""INSERT INTO {audit_table_name} (table_name, record_id, operation_type, changed_at, previous_values, new_values) 
                   VALUES ({','.join(values)});"""
    
    def __get_row_version_json_object_sql(self, table_name, row_version):
        col_names = self._db_client.get_table_column_names(table_name)
        json_object_args = []
        for cn in col_names:
            json_object_args.append(f"'{cn}'")
            json_object_args.append(f"{row_version}.{cn}")
        return f"json_object({','.join(json_object_args)})"





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tables.table_with_header import TableWithHeader
from google.protobuf import json_format
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.tools.tools import Tools
from holado_protobuf.ipc.protobuf.types.google.protobuf import Duration,\
    Timestamp
from holado.common.handlers.enums import AutoNumber

logger = logging.getLogger(__name__)


class SortOrder(AutoNumber):
    """
    Sort order.
    """
    
    Definition = ()
    Alphabetic = ()
    


class ProtobufConverter(object):
    """
    Manage the conversion of Protobuf types (in given ProtobufMessages) to many types.
    """
    def __init__(self): 
        self.__protobuf_messages = None
    
    def initialize(self, protobuf_messages): 
        self.__protobuf_messages = protobuf_messages
    
    def create_table_with_protobuf_fields_as_columns(self, list_proto_obj, recursive=False, uncollapse_repeated=False, with_unset=True):
        res = TableWithHeader()
    
        if len(list_proto_obj) > 0:
            # Set table header
            self.__fill_table_header_with_protobuf_fields(res, list_proto_obj[0], recursive=recursive, uncollapse_repeated=uncollapse_repeated, with_unset=with_unset)
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Result table - set header {res.header.represent(0)}")
            
            # Add table rows
            self.__fill_table_rows_with_protobuf_fields(res, list_proto_obj, uncollapse_repeated=uncollapse_repeated)
            
        return res
    
    def __fill_table_header_with_protobuf_fields(self, res_table, obj, recursive=False, uncollapse_repeated=False, with_unset=True):
        attribute_names = self.__protobuf_messages.get_message_field_names(obj, recursive=recursive, uncollapse_repeated=uncollapse_repeated, with_unset=with_unset)
        res_table.header.add_cells_from_contents(cells_content=attribute_names)
    
    def __fill_table_rows_with_protobuf_fields(self, res_table, list_obj, uncollapse_repeated=False):
        for obj in list_obj:
            if uncollapse_repeated:
                values_by_name = {cn: self.__protobuf_messages.get_object_field_values(obj, cn) for cn in res_table.get_column_names()}
                
                names_uncollapsed = [cn for cn in values_by_name if len(values_by_name[cn]) > 1]
                
                if len(names_uncollapsed) == 0:
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"Result table - add row with {values_by_name}")
                    values_by_name = {cn: values_by_name[cn][0] for cn in values_by_name}
                    res_table.add_row(contents_by_colname=values_by_name)
                elif len(names_uncollapsed) == 1:
                    name_uncollapsed = names_uncollapsed[0]
                    values_uncollapsed = values_by_name.pop(name_uncollapsed)
                    values_by_name = {cn: values_by_name[cn][0] for cn in values_by_name}
                    for value in values_uncollapsed:
                        vbn = dict(values_by_name)
                        vbn[name_uncollapsed] = value
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"Result table - add row with {vbn}")
                        res_table.add_row(contents_by_colname=vbn)
                else:
                    raise TechnicalException("Uncollapse two different repeated fields in the same message is not managed.")
            else:
                values_by_name = {cn: self.__protobuf_messages.get_object_field_value(obj, cn) for cn in res_table.get_column_names()}
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Result table - add row with {values_by_name}")
                res_table.add_row(contents_by_colname=values_by_name)
    
    def convert_protobuf_object_to_json_object(self, proto_obj):
        if isinstance(proto_obj, Timestamp.protobuf_class()) or isinstance(proto_obj, Duration.protobuf_class()):
            # Note: currently, Python doesn't support nanoseconds in datetime, thus it is replaced by string format managed in Protobuf types when available
            return proto_obj.ToJsonString()
        elif self.__protobuf_messages.is_object_repeated(proto_obj):
            return [self.convert_protobuf_object_to_json_object(fv) for fv in proto_obj]
        elif self.__protobuf_messages.is_object_map(proto_obj):
            return dict(proto_obj)
        elif self.__protobuf_messages.is_object_message(proto_obj):
            # Note: following commented code could be an alternative, but it doesn't convert correctly some protobuf types to json
            #     res_str = json_format.MessageToJson(proto_obj)
            #     return json.loads(res_str)
            
            res = {}
            attr_names = self.__protobuf_messages.get_object_field_names(proto_obj, with_unset=False)
            for attr_name in attr_names:
                # Manage special fields
                if hasattr(proto_obj, attr_name):
                    attr_value = getattr(proto_obj, attr_name)
                    if (isinstance(attr_value, Timestamp.protobuf_class()) or isinstance(attr_value, Duration.protobuf_class())
                        or self.__protobuf_messages.is_object_repeated(attr_value) or self.__protobuf_messages.is_object_map(attr_value)
                        or self.__protobuf_messages.is_object_message(attr_value)
                        ):
                        res[attr_name] = self.convert_protobuf_object_to_json_object(attr_value)
                        continue
                    
                # Default
                attr_value = self.__protobuf_messages.get_object_field_value(proto_obj, attr_name)
                if not Converter.is_primitive(attr_value):
                    attr_value = json_format.MessageToDict(attr_value)
                res[attr_name] = attr_value
            return res
        elif Converter.is_primitive(proto_obj):
            return proto_obj
        else:
            raise TechnicalException(f"Unmanaged protobuf object of type '{type(proto_obj)}'")
    
    def convert_protobuf_object_to_name_value_table(self, proto_obj, recursive=False, uncollapse_repeated=False, with_unset=True, sort_order=SortOrder.Definition):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Protobuf object {id(proto_obj)} to Name/Value table")
        res = TableWithHeader()
        res.header.add_cells_from_contents(["Name", "Value"])
        
        attr_names = self.__protobuf_messages.get_object_field_names(proto_obj, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=True, with_unset=with_unset, sort_order=sort_order)
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Protobuf object {id(proto_obj)} to Name/Value table: field names = {attr_names}")
            
        for attr_name in attr_names:
            attr_val = self.__protobuf_messages.get_object_field_value(proto_obj, attr_name)
            row = res.add_row(cells_content=(attr_name, attr_val))
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Converting Protobuf object {id(proto_obj)} to Name/Value table: added row {row}")
            
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Convert Protobuf object {id(proto_obj)} to Name/Value table =>\n{res.represent(indent=4)}")
        return res


    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.finders.tools.find_builder import FindBuilder
from holado_core.common.actors.actions import FindAllAction, FindAction, Action
from holado_core.common.actors.actor import Actor

logger = logging.getLogger(__name__)



class FindActor(Actor):
    """ Base class for find actor.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
        self.__find_builder = None
        
    def initialize(self):
        """
        Initialize actor
        """
        self.__find_builder = self.default_find_builder
    
    @property
    def default_find_builder(self):
        """
        @return New default find builder for this inspector
        """
        return FindBuilder()
    
    @property
    def find_builder(self):
        """
        @return Find builder defined for this inspector
        """
        return self.__find_builder
    
    def act_find(self, finder, find_type, redo=True):
        """
        @param finder Finder
        @param find_type Find type
        @return Action that find element with finder
        """
        res = FindAction(None, self.find_builder, finder=finder, find_type=find_type)
        return self._act_with_redo(res) if redo else res
    
    def act_find_all(self, finder, find_type, redo=True):
        """
        @param finder Finder
        @param find_type Find type
        @return Action that find all elements with finder
        """
        res = FindAllAction(None, self.find_builder, finder=finder, find_type=find_type)
        return self._act_with_redo(res) if redo else res
    
    def act_number_of_occurrences(self, finder=None, find_type=None):
        """
        @return Number of occurrences
        """
        if finder is None:
            def func(elements):
                return len(elements)
            return Action("number of occurences", func)
        else:
            return self.act_then(
                self.act_find_all(finder, find_type),
                self.act_number_of_occurrences() )
    
    def _act_with_redo(self, action):
        """
        Manage redo for given action
        @param action Action
        @return New action managing redo
        """
        if isinstance(action, FindAction) or isinstance(action, FindAllAction):
            def func(container, candidates, find_context, find_parameters):
                if find_parameters.redo:
                    new_parameters = find_parameters.with_redo(False)
                    redo = self._get_redo(action, container, candidates, find_context, new_parameters)
                    redo.ignore_all(find_parameters.redo_ignored_exceptions)
                    return redo.execute()
                else:
                    return action.execute(container, candidates, find_context, find_parameters)
                
            return type(action)(f"{action.name} with redo", self.find_builder, func)
        else:
            return super()._act_with_redo(action)
    
    
    
    
    


from builtins import super
import logging
from holado_test.common.context.test_session_context import BehaveSessionContext

logger = logging.getLogger(__name__)


class TSSessionContext(BehaveSessionContext):
    
    def __init__(self):
        super().__init__("TSSession")
        
    def configure(self, session_kwargs=None):
        logger.info("Configuring TSSessionContext")
        
        super().configure(session_kwargs)
        
        # Override default registered modules
         
        from common.tools.path_manager import TSPathManager  # @UnresolvedImport
        self.services.register_service_type("path_manager", TSPathManager,
                            lambda m: m.initialize(),
                            raise_if_exist=False )
        
        
        # Register new modules
        
        from config.config_manager import TSConfigManager  # @UnresolvedImport
        self.services.register_service_type("config_manager", TSConfigManager,
                            lambda m: m.initialize(lambda: self.path_manager) )
        
        
    def initialize(self, session_kwargs=None):
        if session_kwargs is None:
            session_kwargs = {}
        raise_if_not_exist = session_kwargs.get("raise_if_not_exist", True)
        
        # Call default initialization
        super().initialize(session_kwargs)
        
        # Initialize testing solution
        
        
        
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
import logging
import re
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_scripting.common.tools.variable_manager import VariableManager
from holado_core.common.tools.tools import Tools
from holado_value.common.tables.value_table import ValueTable
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.tables.table import Table
from holado_core.common.tables.table_row import TableRow
from holado_core.common.tables.table_cell import TableCell
from holado_value.common.tables.value_table_row import ValueTableRow
from holado_value.common.tables.value_table_cell import ValueTableCell
from holado_value.common.tools.value_types import ValueTypes
from holado_value.common.tables.comparators.table_2_value_table_with_header_comparator import Table2ValueTable_WithHeaderComparator
from holado_value.common.tables.comparators.table_2_value_table_comparator import Table2ValueTable_Comparator
from holado_core.common.tables.table_manager import TableManager
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.holado_config import Config
from holado_python.standard_library.typing import Typing
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado.common.handlers.undefined import undefined_argument

logger = logging.getLogger(__name__)



class SRE(object):
    """
    Step RegExp
    All registered step parameter types are added in this class as public attributes,
    with type name as attribute name, and pattern as associated value
    """
    
    @staticmethod
    def all_as_dict():
        return Typing.get_object_attribute_values_by_name(SRE)


#TODO: make it a service in SessionContext
class StepTools(object):
    
    __registered_types = {}
    
    @staticmethod
    def _get_scenario_context():
        return SessionContext.instance().get_scenario_context()

    @staticmethod
    def _get_expression_evaluator():
        return StepTools._get_scenario_context().get_expression_evaluator()

    @staticmethod
    def _get_text_interpreter():
        return StepTools._get_scenario_context().get_text_interpreter()

    @staticmethod
    def _get_variable_manager() -> VariableManager:
        return StepTools._get_scenario_context().get_variable_manager()

    @staticmethod
    def unescape_string(text):
        #res = bytes(text, "utf-8").decode("unicode_escape")
        res = text.encode('latin-1').decode('unicode-escape')
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("unescaping string [{}] -> [{}]".format(text, res))
        return res
    
    @staticmethod
    def extract_string_value(text_param):
        return StepTools._get_expression_evaluator().extract_string_value(text_param)
    
    @staticmethod
    def evaluate_string_parameter(text_param, do_interpret=True, log_level=logging.DEBUG):
        eval_params = EvaluateParameters.default().with_interpret(do_interpret).with_eval(False).with_result_type(str)
        _, res = StepTools._get_expression_evaluator().evaluate_expression(text_param, eval_params=eval_params, log_level=Tools.do_log_level(log_level, logging.TRACE))  # @UndefinedVariable
        if Tools.do_log_if_objects_are_different(logger, log_level, text_param, res):
            logger.log(log_level, "evaluate_string_parameter: [{}] -> [{}]".format(text_param, res))
        return res
    
    @staticmethod
    def evaluate_variable_name(text_param, log_level=logging.DEBUG):
        if text_param is None:
            return None
        
        eval_params = EvaluateParameters.default_without_eval(False).with_result_type(str)
        res = StepTools._get_text_interpreter().interpret(text_param, eval_params=eval_params, log_level=Tools.do_log_level(log_level, logging.TRACE))  # @UndefinedVariable
        if Tools.do_log_if_objects_are_different(logger, log_level, text_param, res):
            logger.log(log_level, f"evaluate_variable_name: [{text_param}] -> [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
    
    @staticmethod
    def evaluate_variable_value(text_param, log_level=logging.DEBUG):
        if text_param is None:
            return None
        
        eval_params = EvaluateParameters.default_without_eval(False)
        _, res = StepTools._get_expression_evaluator().evaluate_expression(text_param, eval_params=eval_params, log_level=Tools.do_log_level(log_level, logging.TRACE))  # @UndefinedVariable
        if Tools.do_log_if_objects_are_different(logger, log_level, text_param, res):
            logger.log(log_level, f"evaluate_variable_name: [{text_param}] -> [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
    
    @staticmethod
    def evaluate_scenario_parameter(text_param, log_level=logging.DEBUG):
        if text_param is None:
            return None
        
        eval_params = EvaluateParameters.default().with_raise_on_eval_error(False)
        _, res = StepTools._get_expression_evaluator().evaluate_expression(text_param, eval_params=eval_params, log_level=Tools.do_log_level(log_level, logging.TRACE))  # @UndefinedVariable
        if Tools.do_log_if_objects_are_different(logger, log_level, text_param, res):
            logger.log(log_level, f"evaluate_scenario_parameter: [{text_param}] -> [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
 
    @staticmethod
    def evaluate_list_scenario_parameter(text_param, param_name, log_level=logging.DEBUG):
        res = StepTools.evaluate_scenario_parameter(text_param, log_level=log_level)
        if isinstance(res, str):
            regex = re.compile(r"^\s*\[(.*)\]\s*$")
            m = regex.match(res)
            if m:
                param_names = m.group(1).split(',')
                res = [StepTools.evaluate_scenario_parameter(pn.strip()) for pn in param_names]
            else:
                raise FunctionalException(f"Parameter '{param_name}' must be specified in a list format (ex: \"['Name1', 'Name2',...]\")")
        if Tools.do_log_if_objects_are_different(logger, log_level, text_param, res):
            logger.log(log_level, f"evaluate_list_scenario_parameter for parameter '{param_name}': {text_param} -> {res}")
        return res
    
    @classmethod
    def replace_variable_names(cls, text, log_level=logging.DEBUG):
        res = text
        if res is None:
            return None
        
        pattern_var = re.compile(SRE.VariableName)
        pos = 0
        m = pattern_var.search(res, pos)
        while m:
            var_name = cls._get_variable_manager().evaluate_variable_name(m.group())
            if cls._get_variable_manager().exists_variable(var_name):
                res = res[:m.start()] + var_name + res[m.end():]
                pos = m.start() + len(var_name)
            else:
                pos = m.end()
            m = pattern_var.search(res, pos)
        
        if res != text:
            if Tools.do_log(logger, log_level):
                logger.log(log_level, f"replace_variable_names: [{text}] -> [{res}]")
        return res
    
    @classmethod    
    def get_step_multiline_text(cls, *args, **kwargs):
        """
        Get the step multiline text interpreted.
        Note: parameter context can be the step context or the step object.
        """
        raise NotImplementedError
        
    
    @classmethod    
    def get_step_table(cls, *args, **kwargs):
        """
        Get the step table interpreted.
        Note: parameter context can be the step context or the step object.
        """
        raise NotImplementedError
        
    
    @classmethod
    def convert_step_table_2_value_table(cls, table, do_eval_once=True):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting step table to scenario table (step table = {table})")

        res = ValueTable()
        
        # Set body
        previous_row = None
        for row in cls._get_table_rows(table):
            srow = cls._convert_step_table_row_2_value_table_row(row, previous_row, do_eval_once=do_eval_once)
            res.add_row(row=srow)
            previous_row = srow
            
        # Manage one cell table referencing a variable containing a table
        res = cls._extract_inner_table_if_present(res, do_eval_once=do_eval_once)
            
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting step table to scenario table (step table = {table}) => {res}")
        return res
        
    @classmethod        
    def convert_step_table_2_value_table_with_header(cls, table, do_eval_once=True) -> ValueTableWithHeader :
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting step table to scenario table with header (step table = {table})")
        if table is None:
            raise FunctionalException("This step needs a table")

        res = ValueTableWithHeader()
        
        # Set header
        res.header = cls._convert_step_table_header_2_table_header(cls._get_table_header(table))
        
        # Set body
        previous_row = None
        for row in cls._get_table_rows(table):
            srow = cls._convert_step_table_row_2_value_table_row(row, previous_row, do_eval_once=do_eval_once)
            res.add_row(row=srow)
            previous_row = srow
            
        # Manage one cell table referencing a variable containing a table
        res = cls._extract_inner_table_if_present(res, do_eval_once=do_eval_once)
            
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting step table to scenario table with header (step table = {table}) => {res}")
        return res

    @classmethod
    def _extract_inner_table_if_present(cls, table, do_eval_once=True):
        res = table
        if isinstance(table, TableWithHeader):
            if res.nb_columns == 1 and res.nb_rows == 0:
                cell_content = res.header.get_cell(0).content
                if cls._get_variable_manager().exists_variable(variable_name=cell_content):
                    value = cls._get_variable_manager().get_variable_value(cell_content)
                    if ValueTableManager.is_value_table(value):
                        res = value
                    elif isinstance(value, TableWithHeader):
                        res = ValueTableConverter.convert_table_2_value_table(value, do_eval_once=do_eval_once)
                    elif isinstance(value, Table):
                        raise FunctionalException(f"The table in variable '{cell_content}' is expected to be with header")
        elif isinstance(table, Table):        
            if res.nb_columns == 1 and res.nb_rows == 1:
                value = res.get_row(0).get_cell(0).value
                if ValueTableManager.is_value_table(value):
                    res = value
                elif isinstance(value, Table):
                    res = ValueTableConverter.convert_table_2_value_table(value, do_eval_once=do_eval_once)
                    
        if res != table and Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Extracted inner table: {table} => {res}")
        return res
            
    @classmethod
    def _convert_step_table_header_2_table_header(cls, header):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("Converting step table header to table header")

        res = TableRow()
        
        for hcell in header:
            cell = TableCell(hcell)
            
            # TODO: manage merged header cells
            
            res.add_cell(cell=cell)
           
        return res
        
    @classmethod
    def _convert_step_table_row_2_value_table_row(cls, row, previous_row, do_eval_once=True):  # @UnusedVariable
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("Converting behave table row to scenario table row")
        
        res = ValueTableRow()
        
        for cell in row:
            scell = ValueTableCell(cell, do_eval_once=do_eval_once)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Converting cell [{}] -> scenario table cell (content=[{}] ; value_type='{}')".format(cell, scell.content, scell.value_type.name))
            
            # TODO: manage merged cells
            
            res.add_cell(cell=scell)
           
        return res
    

    @classmethod
    def has_table_header(cls, table):
        """
        Return True if scenario table has a header (for HolAdo framework).
        For HolAdo framework, table header names are all texts without quotes.
        """
        internal_header = cls._get_table_header(table)
        if internal_header is None:
            return False
        
        for cell in internal_header:
            scell = ValueTableCell(cell)
            if not scell.content_type == ValueTypes.Symbol:
                return False
        
        return True

    @classmethod
    def _has_table_header(cls, table):
        """Return True if scenario table has a header (for internal tool interpreting scenario text)."""
        return cls._get_table_header(table) is not None
    
    @classmethod
    def _is_scenario_step_table(cls, table):
        """Return if a table is a scenario step table."""
        raise NotImplementedError()

    @classmethod
    def _get_table_header(cls, table):
        """Return iterable object on scenario table header (for internal tool interpreting scenario text)."""
        raise NotImplementedError()

    @classmethod
    def _get_table_rows(cls, table):
        """Return iterable object on scenario table rows (for internal tool interpreting scenario text)."""
        raise NotImplementedError()
        
    @classmethod
    def represent_step_table(cls, table, indent=0):
        res_list = []
        
        header = cls._get_table_header(table)
        if header:
            res_list.append("| " + " | ".join(header) + " |")
        
        for row in cls._get_table_rows(table):
            res_list.append("| " + " | ".join(row) + " |")
            
        return Tools.indent_string(indent, "\n".join(res_list))
        

    @classmethod
    def then_table_is(cls, table_obtained, table_expected, raise_exception=True):
        is_with_header = TableManager.is_table_with_header(table_obtained)
        if cls._is_scenario_step_table(table_expected):
            if is_with_header:
                table_expected = cls.convert_step_table_2_value_table_with_header(table_expected)
            else:
                table_expected = cls.convert_step_table_2_value_table(table_expected)
        
        if is_with_header:
            comparator = Table2ValueTable_WithHeaderComparator()
        else:
            comparator = Table2ValueTable_Comparator()
        try:
            return comparator.equals(table_obtained, table_expected, is_obtained_vs_expected=True, raise_exception=raise_exception)
        except FunctionalException as exc:
            raise FunctionalException(f"Tables are different (obtained = table 1 ; expected = table 2):\n{Tools.indent_string(4, exc.message)}") from exc

    @classmethod
    def then_table_contains(cls, table_obtained, table_expected, raise_exception=True):
        is_with_header = TableManager.is_table_with_header(table_obtained)
        if cls._is_scenario_step_table(table_expected):
            if is_with_header:
                table_expected = cls.convert_step_table_2_value_table_with_header(table_expected)
            else:
                table_expected = cls.convert_step_table_2_value_table(table_expected)
        
        if is_with_header:
            comparator = Table2ValueTable_WithHeaderComparator()
        else:
            comparator = Table2ValueTable_Comparator()
            
        try:
            return comparator.contains_rows(table_obtained, table_expected, is_obtained_vs_expected=True, raise_exception=raise_exception)
        except FunctionalException as exc:
            raise FunctionalException(f"Obtained table doesn't contain expected table (obtained = table 1 ; expected = table 2):\n{Tools.indent_string(4, exc.message)}") from exc

    @classmethod
    def then_table_doesnt_contain(cls, table_obtained, table_expected, raise_exception=True):
        is_with_header = TableManager.is_table_with_header(table_obtained)
        if cls._is_scenario_step_table(table_expected):
            if is_with_header:
                table_expected = cls.convert_step_table_2_value_table_with_header(table_expected)
            else:
                table_expected = cls.convert_step_table_2_value_table(table_expected)
        
        if is_with_header:
            comparator = Table2ValueTable_WithHeaderComparator()
        else:
            comparator = Table2ValueTable_Comparator()
            
        try:
            return comparator.doesnt_contain_rows(table_obtained, table_expected, is_obtained_vs_expected=True, raise_exception=raise_exception)
        except FunctionalException as exc:
            raise FunctionalException(f"Obtained table contains at least a row of expected table (obtained = table 1 ; expected = table 2):\n{Tools.indent_string(4, exc.message)}") from exc
    
    @classmethod
    def format_step(cls, step_str, keyword=None, table=None, text=None):
        step_str = step_str.strip()
        
        if keyword is not None:
            res = f"{keyword} {step_str}"
        else:
            res = step_str
            
        if table is not None:
            if isinstance(table, Table):
                rendered_table = table.represent(indent=4)
            else:
                raise TechnicalException(f"Unmanaged table of type '{Typing.get_object_class_fullname(table)}'")
            res = u"{res}\n{table}".format(res=res, table=rendered_table)
        elif text is not None:
            rendered_text = text.replace(u'"""', u'\\"\\"\\"')
            rendered_text = Tools.indent_string(4, u'"""\n' + rendered_text + '\n"""\n')
            res = u"{res}\n{text}".format(res=res, text=rendered_text)
            
        return res
        
    @classmethod
    def format_step_short_description(cls, step, step_number, step_context=None, dt_ref=None, has_failed=undefined_argument):
        raise NotImplementedError()
    
    @classmethod
    def format_steps_with(cls, steps, format_with_list):
        res = steps
        for form in format_with_list:
            _, form_eval = cls._get_expression_evaluator().evaluate_expression(form, log_level=logging.TRACE)  # @UndefinedVariable
            if form_eval == form:
                raise FunctionalException(f"Not able to evaluation expression [{form}]. Is missing its initialization ?")
            res = res.replace(f"${{{form}}}", str(form_eval))
        return res
    
    @classmethod
    def format_steps_as_parameter(cls, steps, indent):
        return "\n" + Tools.indent_string(indent, steps+'\n')
    
    @classmethod
    def get_step_description(cls, step):
        raise NotImplementedError()
    
    @classmethod
    def get_step_error_message(cls, step):
        raise NotImplementedError()
    
    @classmethod
    def get_step_error(cls, step):
        raise NotImplementedError()
    
    @classmethod    
    def register_type(cls, type_name, pattern, eval_func, **eval_func_kwargs):
        # Register type in behave
        def func(text):
            return eval_func(text, **eval_func_kwargs)
        func.pattern = pattern
        
        # Add type in SRE class
        setattr(SRE, type_name, pattern)
        
        # Store type as registered
        new_type = {'pattern':pattern, 'function':func}
        if type_name in cls.__registered_types:
            logger.warning(f"Overriding step parameter type '{type_name}': {cls.__registered_types[type_name]['pattern']} -> {new_type['pattern']}")
            import traceback
            logging.warning("".join(traceback.format_list(traceback.extract_stack())))
        cls.__registered_types[type_name] = new_type
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Registered step parameter type '{type_name}'")
        
    @classmethod    
    def get_registered_type_names(cls):
        return list(cls.__registered_types)
        
    @classmethod    
    def get_registered_type_pattern(cls, type_name):
        if type_name not in cls.__registered_types:
            raise TechnicalException(f"Type '{type_name}' is not registered")
        return cls.__registered_types[type_name]['pattern']
        
    @classmethod    
    def get_registered_type_function(cls, type_name):
        if type_name not in cls.__registered_types:
            raise TechnicalException(f"Type '{type_name}' is not registered")
        return cls.__registered_types[type_name]['function']
        
    @classmethod    
    def register_default_types(cls):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Registering step parameter types...")
        ##### Old implementation:
        # __step_regex = {
        #     'variable name': r"[\w${}]+",
        #     'variable expression': r"(?:(?! = | by )[^=])+",
        #     'any parameter': r"(?:[^' ]+|.*)",
        #     'float parameter': r"(?:[^' ]+|\d+(?:\.\d*)?)",
        #     'int parameter': r"(?:[^' ]+|\d+)",
        #     'str parameter': r"(?:[^' ]+|'[^']*'(?:%|#)?)",
        #     'str parameter with quotes': r"(?:[^' ]+|'.*'(?:%|#)?)",
        #     }
        
        regex_suffix = f"(?:{Config.DYNAMIC_SYMBOL}|{Config.THREAD_DYNAMIC_SYMBOL}|{Config.UNIQUE_SYMBOL})?"
        regex_suffix_chars = f"{Config.DYNAMIC_SYMBOL}{Config.THREAD_DYNAMIC_SYMBOL}{Config.UNIQUE_SYMBOL}"
        
        # Variable types
        cls.register_type('VariableName', 
                          r"[\w$\{{\}}]+{suffix}".format(suffix=regex_suffix), 
                          StepTools.evaluate_variable_name)
        # cls.register_type('VariableExpression', r"(?:(?! = | by )[^=])+", StepTools.evaluate_variable_name)
        cls.register_type('VariableExpression', 
                          r"[\w\-\+$\{{\}}\[\]\(\)\.:,'{suffix_chars}]+".format(suffix_chars=regex_suffix_chars), 
                          StepTools.evaluate_variable_name)
        cls.register_type('Variable', 
                          cls.get_registered_type_pattern('VariableExpression'), 
                          StepTools.evaluate_variable_name)
        
        cls.register_type('Any', 
                          r".*", 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('AnyLazy', 
                          r".*?", 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('Float', 
                          r"{Variable}|(?:-)?\d+(?:\.\d*)?".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('Int', 
                          r"{Variable}|(?:-)?\d+".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('Boolean', 
                          r"{Variable}|True|False".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_scenario_parameter)
        
        # String types
        cls.register_type('Bytes', 
                          r"{Variable}|b'[^']*'".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('Str', 
                          r"{Variable}|r?'[^']*'{suffix}".format(Variable=cls.get_registered_type_pattern('Variable'), suffix=regex_suffix), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('RawStr', 
                          r"{Variable}|'.*'{suffix}".format(Variable=cls.get_registered_type_pattern('Variable'), suffix=regex_suffix), 
                          StepTools.evaluate_string_parameter)
        cls.register_type('AnyStr', 
                          r"{Variable}|'.*'{suffix}".format(Variable=cls.get_registered_type_pattern('Variable'), suffix=regex_suffix), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('Code', 
                          r"[^ ]+", 
                          StepTools.evaluate_scenario_parameter)
        
        # Generic types
        cls.register_type('Enum', 
                          r"{Variable}|\w+".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('Dict', 
                          r"{Variable}|\{{.*\}}".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_scenario_parameter)
        cls.register_type('List', 
                          r"{Variable}|\[.*\]".format(Variable=cls.get_registered_type_pattern('Variable')), 
                          StepTools.evaluate_list_scenario_parameter)


